# -*- coding: utf-8 -*-

# Copyright OpenStack Foundation and OpenDev Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.  You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function

from . import hooks

import argparse
import configparser
import datetime
import getpass
import json
import os
import re
import shlex
import stat
import subprocess
import sys
import textwrap
from urllib.parse import quote_plus
from urllib.parse import urlencode
from urllib.parse import urljoin
from urllib.parse import urlparse

import requests

try:
    import importlib.metadata as importlib_metadata
    pkg_resources = None
except ImportError:
    # Pre-py38
    importlib_metadata = None
    import pkg_resources


VERBOSE = False
UPDATE = False
LOCAL_MODE = 'GITREVIEW_LOCAL_MODE' in os.environ
CONFIGDIR = os.path.expanduser("~/.config/git-review")
GLOBAL_CONFIG = "/etc/git-review/git-review.conf"
USER_CONFIG = os.path.join(CONFIGDIR, "git-review.conf")
DEFAULTS = dict(scheme='ssh', hostname=False, port=None, project=False,
                branch='master', remote="gerrit", rebase="1",
                track="0", usepushurl="0", notopic=True, branchauthor="name")
LOCAL_GIT_VERSION = (0, 0, 0)
COPYRIGHT = """\
Copyright OpenStack Foundation and OpenDev Contributors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied.

See the License for the specific language governing permissions and
limitations under the License.
"""

# Bump this if a feature of a more recent git is unconditionally
# required
MINIMUM_GIT_VERSION = (2, 10, 0)


_branch_name = None
_has_color = None
_use_color = None
_orig_head = None
_rewrites = None
_rewrites_push = None
BRANCHAUTHOR = None


class colors(object):
    yellow = '\033[33m'
    green = '\033[92m'
    reset = '\033[0m'
    blue = '\033[36m'


class GitReviewException(Exception):
    EXIT_CODE = 1


class CommandFailed(GitReviewException):

    def __init__(self, *args):
        Exception.__init__(self, *args)
        (self.rc, self.output, self.argv, self.envp) = args
        self.quickmsg = dict([
            ("argv", " ".join(self.argv)),
            ("rc", self.rc),
            ("output", self.output)])

    def __str__(self):
        return """
The following command failed with exit code %(rc)d
    "%(argv)s"
-----------------------
%(output)s
-----------------------""" % self.quickmsg


class ChangeSetException(GitReviewException):

    def __init__(self, e):
        GitReviewException.__init__(self)
        self.e = str(e)

    def __str__(self):
        return self.e


def printwrap(unwrapped):
    print('\n'.join(textwrap.wrap(unwrapped)))


def warn(warning):
    printwrap("WARNING: %s" % warning)


def parse_review_number(review):
    parts = review.split(',')
    if len(parts) < 2:
        parts.append(None)
    return parts


def build_review_number(review, patchset):
    if patchset is not None:
        return '%s,%s' % (review, patchset)
    return review


def run_command_status(*argv, **kwargs):
    if VERBOSE:
        print(datetime.datetime.now(), "Running:", " ".join(argv))
    if len(argv) == 1:
        argv = shlex.split(str(argv[0]))
    stdin = kwargs.pop('stdin', None)
    newenv = os.environ.copy()
    newenv['LANG'] = 'C'
    newenv['LANGUAGE'] = 'C'
    newenv.update(kwargs)
    p = subprocess.Popen(argv,
                         stdin=subprocess.PIPE if stdin else None,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT,
                         env=newenv, universal_newlines=True)
    (out, nothing) = p.communicate(stdin)
    return (p.returncode, out.strip())


def run_command(*argv, **kwargs):
    (rc, output) = run_command_status(*argv, **kwargs)
    return output


def run_command_exc(klazz, *argv, **env):
    """Run command *argv, on failure raise klazz

    klazz should be derived from CommandFailed
    """
    (rc, output) = run_command_status(*argv, **env)
    if rc:
        raise klazz(rc, output, argv, env)
    return output


def git_credentials(url):
    """Return credentials using git credential or None."""
    cmd = 'git', 'credential', 'fill'
    stdin = 'url=%s' % url
    rc, out = run_command_status(*cmd, stdin=stdin)
    if rc:
        return None
    data = dict(line.split('=', 1) for line in out.splitlines())
    return data['username'], data['password']


def http_code_2_return_code(code):
    """Tranform http status code to system return code."""
    return (code - 301) % 255 + 1


def run_http_exc(klazz, url, **env):
    """Run http GET request url, on failure raise klazz

    klazz should be derived from CommandFailed
    """
    if url.startswith("https://") and "verify" not in env:
        if "GIT_SSL_NO_VERIFY" in os.environ:
            env["verify"] = False
        else:
            verify = git_config_get_value("http", "sslVerify", as_bool=True)
            env["verify"] = verify != 'false'

    try:
        res = requests.get(url, **env)
        if res.status_code == 401:
            creds = git_credentials(url)
            if creds:
                env['auth'] = creds
                res = requests.get(url, **env)
    except klazz:
        raise
    except Exception as err:
        raise klazz(255, str(err), ('GET', url), env)
    if not 200 <= res.status_code < 300:
        raise klazz(http_code_2_return_code(res.status_code),
                    res.text, ('GET', url), env)
    return res


def get_version():
    if importlib_metadata:
        return importlib_metadata.version('git-review')
    else:
        requirement = pkg_resources.Requirement.parse('git-review')
        provider = pkg_resources.get_provider(requirement)
        return provider.version


def get_git_version():
    global LOCAL_GIT_VERSION
    output = run_command("git version")
    if "git version" in output:
        try:
            # Git version on Linux is of the form:
            #   git version 2.35.3
            # But on OS X we get:
            #   git version 2.20.1 (Apple Git-117)
            # Keep this as simple as possible by splitting on whitespace
            # and then selecting the 3rd element which should be the version.
            v = output.split()[2]
            LOCAL_GIT_VERSION = tuple(map(int, v.split('.')[:3]))
        except Exception:
            printwrap("Could not determine git version!")
            sys.exit(1)
    if LOCAL_GIT_VERSION < MINIMUM_GIT_VERSION:
        printwrap("Local git version %s < required git version %s" % (
                  '.'.join(map(str, LOCAL_GIT_VERSION)),
                  '.'.join(map(str, MINIMUM_GIT_VERSION))))
        sys.exit(1)


def git_directories():
    """Determine (absolute git work directory path, .git subdirectory path)."""
    cmd = ("git", "rev-parse", "--show-toplevel", "--git-dir")
    out = run_command_exc(GitDirectoriesException, *cmd)
    try:
        return out.splitlines()
    except ValueError:
        raise GitDirectoriesException(0, out, cmd, {})


class GitDirectoriesException(CommandFailed):
    "Cannot determine where .git directory is."
    EXIT_CODE = 70


class CustomScriptException(CommandFailed):
    """Custom script execution failed."""
    EXIT_CODE = 71


def run_custom_script(action):
    """Get status and output of .git/hooks/$action-review or/and
    ~/.config/hooks/$action-review if existing.
    """
    returns = []
    script_file = "%s-review" % (action)
    (top_dir, git_dir) = git_directories()
    paths = [os.path.join(CONFIGDIR, "hooks", script_file),
             os.path.join(git_get_hooks_path(top_dir, git_dir), script_file)]
    for fpath in paths:
        if os.path.isfile(fpath) and os.access(fpath, os.X_OK):
            status, output = run_command_status(fpath)
            returns.append((status, output, fpath))

    for (status, output, path) in returns:
        if status:
            raise CustomScriptException(status, output, [path], {})
        elif output and VERBOSE:
            print("script %s output is:" % (path))
            print(output)


def git_config_get_value(section, option, default=None, as_bool=False):
    """Get config value for section/option."""
    cmd = ["git", "config", "--get", "%s.%s" % (section, option)]
    if as_bool:
        cmd.insert(2, "--bool")
    if LOCAL_MODE:
        __, git_dir = git_directories()
        cmd[2:2] = ['-f', os.path.join(git_dir, 'config')]
    try:
        result = run_command_exc(GitConfigException, *cmd).strip()
        if VERBOSE:
            print(datetime.datetime.now(), "... %s.%s = %s"
                  % (section, option, result))
        return result
    except GitConfigException as exc:
        if exc.rc == 1:
            if VERBOSE and default is not None:
                print(datetime.datetime.now(),
                      "... nothing in git config, returning func parameter:",
                      default)
            return default
        raise


def git_get_hooks_path(top_dir, git_dir):
    """Get the path where we need to store and retrieve Git hooks.

    Normally hooks go into .git/hooks, but users can override with the
    core.hooksPath option. This can either be an absolute path, in
    which case we use it as-is, or a relative path, in which case we
    must interpret it as relative to top_dir.
    """
    hook_dir = os.path.join(git_dir, "hooks")
    hooks_path_option = git_config_get_value('core', 'hooksPath')
    if hooks_path_option:
        if os.path.isabs(hooks_path_option):
            hook_dir = hooks_path_option
        else:
            hook_dir = os.path.join(top_dir, hooks_path_option)
    return hook_dir


class Config(object):

    """Expose as dictionary configuration options."""

    def __init__(self, config_file=None):
        self.config = DEFAULTS.copy()
        filenames = [] if LOCAL_MODE else [GLOBAL_CONFIG, USER_CONFIG]
        if config_file:
            filenames.append(config_file)
        for filename in filenames:
            if os.path.exists(filename):
                if filename != config_file:
                    msg = ("Using global/system git-review config files (%s) "
                           "is deprecated and will be removed in a future "
                           "release")
                    warn(msg % filename)
                self.config.update(load_config_file(filename))

    def __getitem__(self, key):
        """Let 'git config --get' override every Config['key'] access"""
        value = git_config_get_value('gitreview', key)
        if value is None:
            value = self.config[key]
        # "--verbose" doesn't trace *early* invocations; for that you
        # must change the value at the top of this file (*and* pass
        # --verbose)
        if VERBOSE:
            print(datetime.datetime.now(),
                  "Config['%s'] = %s " % (key, value))
        return value


class GitConfigException(CommandFailed):
    """Git config value retrieval failed."""
    EXIT_CODE = 128


class CannotInstallHook(CommandFailed):
    "Problems encountered installing commit-msg hook"
    EXIT_CODE = 2


def set_hooks_commit_msg(remote, target_file):
    """Install the commit message hook if needed."""

    # Create the hooks directory if it's not there already
    hooks_dir = os.path.dirname(target_file)
    if not os.path.isdir(hooks_dir):
        os.mkdir(hooks_dir)

    if not os.path.exists(target_file) or UPDATE:
        remote_url = get_remote_url(remote)
        if not remote_url:
            with open(target_file, 'w') as f:
                f.write(hooks.COMMIT_MSG)
        elif (remote_url.startswith('http://') or
                remote_url.startswith('https://')):
            hook_url = urljoin(remote_url, '/tools/hooks/commit-msg')
            if VERBOSE:
                print("Fetching commit hook from: %s" % hook_url)
            res = run_http_exc(CannotInstallHook, hook_url, stream=True)
            with open(target_file, 'wb') as f:
                for x in res.iter_content(1024):
                    f.write(x)
        else:
            (hostname, username, port, project_name) = \
                parse_gerrit_ssh_params_from_git_url(remote_url)
            if username:
                userhost = "%s@%s" % (username, hostname)
            else:
                userhost = hostname
            # OS independent target file
            scp_target_file = target_file.replace(os.sep, "/")

            # Get scp options
            scp_out = run_command("scp")
            scp_opts = scp_out[scp_out.index("[") + 2:scp_out.index("]")]

            cmd = ["scp", userhost + ":hooks/commit-msg", scp_target_file]
            if port is not None:
                cmd.insert(1, "-P%s" % port)
            # Force scp protocol if the -O option is available
            if "O" in scp_opts:
                cmd.insert(1, "-O")

            if VERBOSE:
                hook_url = 'scp://%s%s/hooks/commit-msg' \
                           % (userhost, (":%s" % port) if port else "")
                print("Fetching commit hook from: %s" % hook_url)
            run_command_exc(CannotInstallHook, *cmd)
        # If there are submodules, the hook needs to be installed into
        # each of them.
        # Here, we don't check for any nonstandard hooks path, because
        # it should be safe to assume that very few users are inclined
        # to set the core.hooksPath option in a submodule checkout.
        run_command_exc(
            CannotInstallHook,
            "git", "submodule", "foreach",
            'cp -p %s "$(git rev-parse --git-dir)/hooks/"' % target_file)

    old_mask = stat.S_IMODE(os.stat(target_file).st_mode)
    # make sure the owner always has read+write+exec perms
    mask = old_mask | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
    # add executable permission for everyone else who has read access
    if mask | stat.S_IRGRP == mask:
        mask |= stat.S_IXGRP
    if mask | stat.S_IROTH == mask:
        mask |= stat.S_IXOTH
    # only call chmod if we need to change the mask
    if mask != old_mask:
        os.chmod(target_file, mask)


def test_remote_url(remote_url):
    """Tests that a possible gerrit remote url works."""
    status, description = run_command_status("git", "push", "--dry-run",
                                             remote_url, "--all")
    if status != 128:
        if VERBOSE:
            print("%s worked. Description: %s" % (remote_url, description))
        return True
    else:
        print("%s did not work. Description: %s" % (remote_url, description))
        return False


def make_remote_url(scheme, username, hostname, port, project):
    """Builds a gerrit remote URL."""
    if port is None and scheme == 'ssh':
        port = 29418
    hostport = '%s:%s' % (hostname, port) if port else hostname
    if username:
        return "%s://%s@%s/%s" % (scheme, username, hostport, project)
    else:
        return "%s://%s/%s" % (scheme, hostport, project)


def add_remote(scheme, hostname, port, project, remote, usepushurl):
    """Adds a gerrit remote."""
    asked_for_username = False

    username = git_config_get_value("gitreview", "username")
    if not username:
        username = getpass.getuser()

    remote_url = make_remote_url(scheme, username, hostname, port, project)
    if VERBOSE:
        print("No remote set, testing %s" % remote_url)
    if not test_remote_url(remote_url):
        print("Could not connect to gerrit.")
        username = input("Enter your gerrit username: ")
        remote_url = make_remote_url(scheme, username, hostname, port, project)
        print("Trying again with %s" % remote_url)
        if not test_remote_url(remote_url):
            raise GerritConnectionException(
                "Could not connect to gerrit at %s" % remote_url
            )
        asked_for_username = True

    if usepushurl:
        cmd = "git remote set-url --push %s %s" % (remote, remote_url)
        print("Adding a git push url to '%s' that maps to:" % remote)
    else:
        cmd = "git remote add -f %s %s" % (remote, remote_url)
        print("Creating a git remote called '%s' that maps to:" % remote)
    print("\t%s" % remote_url)

    (status, remote_output) = run_command_status(cmd)
    if status:
        raise CommandFailed(status, remote_output, cmd, {})

    if asked_for_username:
        print()
        printwrap("This repository is now set up for use with git-review. "
                  "You can set the default username for future repositories "
                  "with:")
        print('  git config --global --add gitreview.username "%s"' % username)
        print()


def populate_rewrites():
    """Populate the global _rewrites and _rewrites_push maps based on the
    output of "git-config".
    """

    cmd = ['git', 'config', '--list']
    out = run_command_exc(CommandFailed, *cmd).strip()

    global _rewrites, _rewrites_push
    _rewrites = {}
    _rewrites_push = {}

    for entry in out.splitlines():
        key, _, value = entry.partition('=')
        key = key.lower()

        if key.startswith('url.') and key.endswith('.insteadof'):
            rewrite = key[len('url.'):-len('.insteadof')]
            if rewrite:
                _rewrites[value] = rewrite
        elif key.startswith('url.') and key.endswith('.pushinsteadof'):
            rewrite = key[len('url.'):-len('.pushinsteadof')]
            if rewrite:
                _rewrites_push[value] = rewrite


def alias_url(url, rewrite_push):
    """Expand a remote URL. Use the global map _rewrites to replace the
    longest match with its equivalent. If rewrite_push is True, try
    _rewrites_push before _rewrites.
    """

    if _rewrites is None:
        populate_rewrites()

    if rewrite_push:
        maps = [_rewrites_push, _rewrites]
    else:
        maps = [_rewrites]

    for rewrites in maps:
        # If Git finds a pushInsteadOf alias, it uses that even if
        # there is a longer insteadOf alias.
        longest = None
        for alias in rewrites:
            if url.startswith(alias) and (
                    longest is None or len(longest) < len(alias)):
                longest = alias

        if longest:
            return url.replace(longest, rewrites[longest])

    return url


def get_remote_url(remote):
    """Retrieve the remote URL. Read the configuration to expand the URL of a
    remote repository taking into account any "url.<base>.insteadOf" or
    "url.<base>.pushInsteadOf" config setting.

    TODO: Replace current code with something like "git ls-remote
    --get-url" after Git grows a version of it that returns the push
    URL rather than the fetch URL.
    """

    # Short-circuit to a false value if the remote passed in is falsey, used
    # for determining commit-msg hook retrieval behavior
    if not remote:
        return remote

    push_url = git_config_get_value('remote.%s' % remote, 'pushurl')
    if push_url is not None:
        # Git rewrites pushurl using insteadOf but not pushInsteadOf.
        push_url = alias_url(push_url, False)
    else:
        url = git_config_get_value('remote.%s' % remote, 'url')
        # Git rewrites url using pushInsteadOf or insteadOf.
        push_url = alias_url(url, True)
    if VERBOSE:
        print("Found origin Push URL:", push_url)
    return push_url


def parse_gerrit_ssh_params_from_git_url(git_url):
    """Parse a given Git "URL" into Gerrit parameters. Git "URLs" are either
    real URLs or SCP-style addresses.
    """

    # The exact code for this in Git itself is a bit obtuse, so just do
    # something sensible and pythonic here instead of copying the exact
    # minutiae from Git.

    # Handle real(ish) URLs
    if "://" in git_url:
        parsed_url = urlparse(git_url)
        path = parsed_url.path

        hostname = parsed_url.netloc
        username = None
        port = parsed_url.port

        # Workaround bug in urlparse on OSX
        if parsed_url.scheme == "ssh" and parsed_url.path[:2] == "//":
            hostname = parsed_url.path[2:].split("/")[0]

        if "@" in hostname:
            (username, _, hostname) = hostname.rpartition("@")
        if ":" in hostname:
            (hostname, port) = hostname.split(":")

        if port is not None:
            port = str(port)

    # Handle SCP-style addresses
    else:
        username = None
        port = None
        (hostname, path) = git_url.split(":", 1)
        if "@" in hostname:
            (username, hostname) = hostname.split("@", 1)

    # Strip leading slash and trailing .git from the path to form the project
    # name.
    project_name = re.sub(r"^/|(\.git$)", "", path)

    return (hostname, username, port, project_name)


def query_reviews(remote_url, project=None, branch=None, change=None,
                  current_patch_set=True, exception=CommandFailed,
                  parse_exc=Exception):
    if remote_url.startswith('http://') or remote_url.startswith('https://'):
        query = query_reviews_over_http
    else:
        query = query_reviews_over_ssh
    return query(remote_url,
                 project=project,
                 branch=branch,
                 change=change,
                 current_patch_set=current_patch_set,
                 exception=exception,
                 parse_exc=parse_exc)


def query_reviews_over_http(remote_url, project=None, branch=None, change=None,
                            current_patch_set=True, exception=CommandFailed,
                            parse_exc=Exception):
    if project:
        # Remove any trailing .git suffixes for project to url comparison
        clean_url = os.path.splitext(remote_url)[0]
        clean_project = os.path.splitext(project)[0]
        if clean_url.endswith(clean_project):
            # Get the "root" url for gerrit by removing the project from the
            # url. For example:
            # https://example.com/foo/project.git gets truncated to
            # https://example.com/foo/ regardless of whether or not none,
            # either, or both of the remote_url or project strings end
            # with .git.
            remote_url = clean_url[:-len(clean_project)]
    url = urljoin(remote_url, 'changes/')
    if change:
        if current_patch_set:
            url += '?q=%s&o=CURRENT_REVISION' % change
        else:
            url += '?q=%s&o=ALL_REVISIONS' % change
    else:
        if project:
            project_name = re.sub(r"^/|(\.git$)", "",
                                  project)
        else:
            project_name = re.sub(r"^/|(\.git$)", "",
                                  urlparse(remote_url).path)
        query = 'project:%s status:open' % project_name
        if branch:
            query += ' branch:%s' % branch
        params = urlencode({'q': query})
        url += '?' + params

    if VERBOSE:
        print("Query gerrit %s" % url)
    request = run_http_exc(exception, url)
    if VERBOSE:
        print(request.text)
    reviews = json.loads(request.text[4:])

    # Reformat output to match ssh output
    try:
        for review in reviews:
            review["number"] = str(review.pop("_number"))
            if "revisions" not in review:
                continue
            patchsets = {}
            for key, revision in review["revisions"].items():
                fetch_value = list(revision["fetch"].values())[0]
                patchset = {"number": str(revision["_number"]),
                            "ref": fetch_value["ref"]}
                patchsets[key] = patchset
            review["patchSets"] = patchsets.values()
            review["currentPatchSet"] = patchsets[review["current_revision"]]
    except Exception as err:
        raise parse_exc(err)

    return reviews


def query_reviews_over_ssh(remote_url, project=None, branch=None, change=None,
                           current_patch_set=True, exception=CommandFailed,
                           parse_exc=Exception):
    (hostname, username, port, project_name) = \
        parse_gerrit_ssh_params_from_git_url(remote_url)

    if change:
        if current_patch_set:
            query = "--current-patch-set change:%s" % change
        else:
            query = "--patch-sets change:%s" % change
    else:
        query = "project:%s status:open" % project_name
        if branch:
            query += ' branch:%s' % branch

    port_data = "p%s" % port if port is not None else ""
    if username is None:
        userhost = hostname
    else:
        userhost = "%s@%s" % (username, hostname)

    if VERBOSE:
        print("Query gerrit %s %s" % (remote_url, query))
    output = run_command_exc(
        exception,
        os.environ.get("GIT_SSH", "ssh"), "-x" + port_data, userhost,
        "gerrit", "query",
        "--format=JSON %s" % query)
    if VERBOSE:
        print(output)

    changes = []
    try:
        for line in output.split("\n"):
            if line[0] == "{":
                try:
                    data = json.loads(line)
                    if "type" not in data:
                        changes.append(data)
                except Exception:
                    if VERBOSE:
                        print(output)
    except Exception as err:
        raise parse_exc(err)
    return changes


def set_color_output(color="auto"):
    global _use_color
    if check_color_support():
        if color == "auto":
            check_use_color_output()
        else:
            _use_color = color == "always"


def check_use_color_output():
    global _use_color
    if _use_color is None:
        if check_color_support():
            # we can support color, now check if we should use it
            stdout = "true" if sys.stdout.isatty() else "false"
            test_command = "git config --get-colorbool color.review " + stdout
            color = run_command(test_command)
            _use_color = color == "true"
        else:
            _use_color = False
    return _use_color


def check_color_support():
    global _has_color
    if _has_color is None:
        test_command = "git log --color=never --oneline HEAD^1..HEAD"
        (status, output) = run_command_status(test_command)
        if status == 0:
            _has_color = True
        else:
            _has_color = False
    return _has_color


def load_config_file(config_file):
    """Load configuration options from a file."""
    configParser = configparser.ConfigParser()
    configParser.read(config_file)
    options = {
        'scheme': 'scheme',
        'hostname': 'host',
        'port': 'port',
        'project': 'project',
        'branch': 'defaultbranch',
        'remote': 'defaultremote',
        'rebase': 'defaultrebase',
        'track': 'track',
        'notopic': 'notopic',
        'usepushurl': 'usepushurl',
        'branchauthor': 'branchauthor',
    }
    config = {}
    for config_key, option_name in options.items():
        if configParser.has_option('gerrit', option_name):
            config[config_key] = configParser.get('gerrit', option_name)
    return config


def update_remote(remote):
    cmd = "git remote update %s" % remote
    (status, output) = run_command_status(cmd)
    if VERBOSE:
        print(output)
    if status != 0:
        print("Problem running '%s'" % cmd)
        if not VERBOSE:
            print(output)
        return False
    return True


def parse_tracking(ref=None):
    """Return tracked (remote, branch) of current HEAD or other named
       branch if tracking remote.
    """
    if ref is None:
        ref = run_command_exc(
            SymbolicRefFailed,
            "git", "symbolic-ref", "-q", "HEAD")
    tracked = run_command_exc(
        ForEachRefFailed,
        "git", "for-each-ref", "--format=%(upstream)", ref)

    # Only on explicitly tracked remote branch do we diverge from default
    if tracked and tracked.startswith('refs/remotes/'):
        return tracked[13:].partition('/')[::2]

    return None, None


def resolve_tracking(remote, branch):
    """Resolve tracked upstream remote/branch if current branch is tracked."""
    tracked_remote, tracked_branch = parse_tracking()
    # tracked_branch will be empty when tracking a local branch
    if tracked_branch:
        if VERBOSE:
            print('Following tracked %s/%s rather than default %s/%s' % (
                  tracked_remote, tracked_branch,
                  remote, branch))
        return tracked_remote, tracked_branch

    return remote, branch


def check_remote(branch, remote, scheme, hostname, port, project,
                 usepushurl=False):
    """Check that a Gerrit Git remote repo exists, if not, set one."""

    if usepushurl:
        push_url = git_config_get_value('remote.%s' % remote, 'pushurl', None)
        if push_url:
            return
    else:
        has_color = check_color_support()
        if has_color:
            color_never = "--color=never"
        else:
            color_never = ""

        if remote in run_command("git remote").split("\n"):

            remotes = run_command("git branch -a %s" % color_never).split("\n")
            for current_remote in remotes:
                remote_string = "remotes/%s/%s" % (remote, branch)
                if (current_remote.strip() == remote_string and not UPDATE):
                    return
            # We have the remote, but aren't set up to fetch. Fix it
            if VERBOSE:
                print("Setting up gerrit branch tracking for better rebasing")
            update_remote(remote)
            return

    if hostname is False or project is False:
        # This means there was no .gitreview file
        printwrap("No '.gitreview' file found in this repository. We don't "
                  "know where your gerrit is.")
        if usepushurl:
            printwrap("Please set the push-url on your origin remote to the "
                      "location of your gerrit server and try again")
        else:
            printwrap("Please manually create a remote named \"%s\" or "
                      "rename the default one and try again." % remote)
        sys.exit(1)

    # Gerrit remote not present, try to add it
    try:
        add_remote(scheme, hostname, port, project, remote, usepushurl)
    except Exception:
        if usepushurl:
            printwrap("We don't know where your gerrit is. Please manually"
                      " add a push-url to the '%s' remote and try again."
                      % remote)
        else:
            printwrap("We don't know where your gerrit is. Please manually"
                      " create a remote named '%s' and try again." % remote)
        raise


def rebase_changes(branch, remote, interactive=True, keep=False):

    global _orig_head

    remote_branch = "remotes/%s/%s" % (remote, branch)

    if not update_remote(remote):
        return False

    # since the value of ORIG_HEAD may not be set by rebase as expected
    # for use in undo_rebase, make sure to save it explicitly
    cmd = "git rev-parse HEAD"
    (status, output) = run_command_status(cmd)
    if status != 0:
        print("Errors running %s" % cmd)
        if interactive:
            print(output)
        return False
    _orig_head = output

    # Avoid trying to do a test rebase if there are uncommitted changes.
    # Either the rebase will fail with a similar message, or if the user
    # has turned on rebase.autostash then the subsequent reset will
    # silently discard those changes.
    cmd = "git diff --ignore-submodules --quiet"
    (status, output) = run_command_status(cmd)
    if status != 0:
        printwrap("You have unstaged changes. Please commit or stash them "
                  "first, and then try again.")
        sys.exit(1)
    cmd = "git diff --cached --ignore-submodules --quiet"
    (status, output) = run_command_status(cmd)
    if status != 0:
        printwrap("You have uncommitted changes. Please commit or stash them "
                  "first, and then try again.")
        sys.exit(1)

    cmd = "git show-ref --quiet --verify refs/%s" % remote_branch
    (status, output) = run_command_status(cmd)
    if status != 0:
        printwrap("The branch '%s' does not exist on the given remote '%s'. "
                  "If these changes are intended to start a new branch, "
                  "re-run with the '-R' option enabled." % (branch, remote))
        sys.exit(1)

    rebase_flag = '--rebase-merges'
    if LOCAL_GIT_VERSION < (2, 18, 0):
        rebase_flag = "--preserve-merges"

    interactive_flag = interactive and '-i' or ''

    cmd = "git rebase %s %s %s" % \
        (rebase_flag, interactive_flag, remote_branch)

    (status, output) = run_command_status(cmd, GIT_EDITOR='true')
    if status != 0:
        print("Errors running %s" % cmd)
        if interactive:
            print(output)
            if keep:
                printwrap("It is likely that your change has a merge "
                          "conflict. You may resolve it in the working tree "
                          "now as described above and then run 'git review' "
                          "again, or if you do not want to resolve it yet "
                          "(note that the change can not merge until the "
                          "conflict is resolved) you may run 'git rebase "
                          "--abort' then 'git review -R' to upload the change "
                          "without rebasing.")
            else:
                printwrap("It is likely that your change has a merge "
                          "conflict, but the result of the test rebase has "
                          "been discarded. You may rebase it yourself and "
                          "review again, or use 'git review -R' to upload the "
                          "change without resolving the conflict (though this "
                          "is discouraged since it probably can't be merged "
                          "in that state), or use the --keep-rebase option to "
                          "leave the incomplete rebase result in your working "
                          "tree.")
        return False
    return True


def abort_rebase():
    cmd = "git rebase --abort"
    (status, output) = run_command_status(cmd)
    if status != 0:
        print("Errors running %s" % cmd)
        print(output)


def undo_rebase():
    global _orig_head
    if not _orig_head:
        return True

    cmd = "git reset --hard %s" % _orig_head
    (status, output) = run_command_status(cmd)
    if status != 0:
        print("Errors running %s" % cmd)
        print(output)
        return False
    return True


def get_branch_name(target_branch):
    global _branch_name
    if _branch_name is not None:
        return _branch_name
    cmd = "git rev-parse --symbolic-full-name --abbrev-ref HEAD"
    _branch_name = run_command(cmd)
    if _branch_name == "HEAD":
        # detached head or no branch found
        _branch_name = target_branch
    return _branch_name


def assert_one_change(remote, branch, yes, have_hook):
    if check_use_color_output():
        use_color = "--color=always"
    else:
        use_color = "--color=never"
    cmd = ("git log %s --decorate --oneline --no-show-signature "
           "HEAD --not --remotes=%s" % (use_color, remote))
    (status, output) = run_command_status(cmd)
    if status != 0:
        print("Had trouble running %s" % cmd)
        print(output)
        sys.exit(1)
    filtered = filter(None, output.split("\n"))
    output_lines = sum(1 for s in filtered)
    if output_lines == 1 and not have_hook:
        printwrap("Your change was committed before the commit hook was "
                  "installed. Amending the commit to add a gerrit change id.")
        run_command("git commit --amend", GIT_EDITOR='true')
    elif output_lines == 0:
        printwrap("The commit HEAD already exists on a branch on the remote. "
                  "If it already exists on %s/%s, then submitting for review "
                  "would be pointless." % (remote, branch))
        print("\n\n"
              "Do you really want to submit this review?")
        try:
            yes_no = input("Type 'yes' to confirm, other to cancel: ")
        except KeyboardInterrupt:
            yes_no = "no"
        if yes_no.lower().strip() != "yes":
            print("Aborting.")
            sys.exit(1)
    elif output_lines > 1:
        if not yes:
            printwrap("You are about to submit multiple commits. This is "
                      "expected if you are submitting a commit that is "
                      "dependent on one or more in-review commits, or if you "
                      "are submitting multiple self-contained but dependent "
                      "changes. Otherwise you should consider squashing your "
                      "changes into one commit before submitting (for "
                      "indivisible changes) or submitting from separate "
                      "branches (for independent changes).")
            print("\nThe outstanding commits are:\n\n%s\n\n"
                  "Do you really want to submit the above commits?" % output)
            try:
                yes_no = input("Type 'yes' to confirm, other to cancel: ")
            except KeyboardInterrupt:
                yes_no = "no"
            if yes_no.lower().strip() != "yes":
                print("Aborting.")
                sys.exit(1)


class CannotQueryOpenChangesets(CommandFailed):
    "Cannot fetch review information from gerrit"
    EXIT_CODE = 32


class CannotParseOpenChangesets(ChangeSetException):
    "Cannot parse JSON review information from gerrit"
    EXIT_CODE = 33


class Review(dict):
    _default_fields = ('branch', 'topic', 'subject')

    def __init__(self, data):
        if 'number' not in data:
            raise TypeError("<Review> requires 'number' key in data")

        super(Review, self).__init__(data)

        # provide default values for some fields
        for field in self._default_fields:
            self[field] = self.get(field, '-')


class ReviewsPrinter(object):
    def __init__(self, with_topic=False):
        if with_topic:
            self.fields = ('number', 'branch', 'topic', 'subject')
            # > is right justify, < is left, field indices for py26
            self.fields_format = [
                u"{0:>{1}}", u"{2:>{3}}", u"{4:>{5}}", u"{6:<{7}}"]
        else:
            self.fields = ('number', 'branch', 'subject')
            # > is right justify, < is left, field indices for py26
            self.fields_format = [u"{0:>{1}}", u"{2:>{3}}", u"{4:<{5}}"]

        self.fields_colors = ("", "", "", "")
        self.color_reset = ""
        if check_use_color_output():
            self.fields_colors = (
                colors.yellow, colors.green, colors.blue, "")
            self.color_reset = colors.reset

        self.reviews = []

    @property
    def fields_width(self):
        return [
            max(len(str(review[field])) for review in self.reviews)
            for field in self.fields[:-1]
        ] + [1]

    def _get_field_format_str(self, field):
        index = self.fields.index(field)
        return (
            self.fields_colors[index] +
            self.fields_format[index] +
            self.color_reset
        )

    def add_review(self, review):
        self.reviews.append(review)

    def _get_fields_format_str(self):
        return "  ".join([
            self._get_field_format_str(field)
            for field in self.fields])

    def print_review(self, review):
        fields_format_str = self._get_fields_format_str()

        formatted_fields = []
        for field, width in zip(self.fields, self.fields_width):
            formatted_fields.extend((
                review[field], width
            ))

        print(fields_format_str.format(*formatted_fields))

    def do_print(self, reviews):

        total_reviews = len(reviews)

        for review in reviews:
            self.print_review(review)

        print("Found %d items for review" % total_reviews)


def list_reviews(remote, project, branch=None, with_topic=False):
    remote_url = get_remote_url(remote)

    reviews = []
    for r in query_reviews(remote_url,
                           project=project,
                           branch=branch,
                           exception=CannotQueryOpenChangesets,
                           parse_exc=CannotParseOpenChangesets):
        reviews.append(Review(r))

    if not reviews:
        print("No pending reviews")
        return

    printer = ReviewsPrinter(with_topic=with_topic)
    for review in reviews:
        printer.add_review(review)

    printer.do_print(reviews)
    return 0


class CannotQueryPatchSet(CommandFailed):
    "Cannot query patchset information"
    EXIT_CODE = 34


class ReviewInformationNotFound(ChangeSetException):
    "Could not fetch review information for change %s"
    EXIT_CODE = 35


class ReviewNotFound(ChangeSetException):
    "Gerrit review %s not found"
    EXIT_CODE = 36


class PatchSetGitFetchFailed(CommandFailed):
    """Cannot fetch patchset contents

Does specified change number belong to this project?
"""
    EXIT_CODE = 37


class PatchSetNotFound(ChangeSetException):
    "Review patchset %s not found"
    EXIT_CODE = 38


class GerritConnectionException(GitReviewException):
    """Problem to establish connection to gerrit."""
    EXIT_CODE = 40


class CheckoutNewBranchFailed(CommandFailed):
    "Cannot checkout to new branch"
    EXIT_CODE = 64


class CheckoutExistingBranchFailed(CommandFailed):
    "Cannot checkout existing branch"
    EXIT_CODE = 65


class ResetKeepFailed(CommandFailed):
    "Failed to reset downloaded branch"
    EXIT_CODE = 66


class SetUpstreamBranchFailed(CommandFailed):
    "Cannot set upstream to remote branch"
    EXIT_CODE = 67


class SymbolicRefFailed(CommandFailed):
    "Cannot find symbolic reference"
    EXIT_CODE = 68


class ForEachRefFailed(CommandFailed):
    "Cannot process symbolic reference"
    EXIT_CODE = 69


class BranchTrackingMismatch(GitReviewException):
    "Branch exists but is tracking unexpected branch"
    EXIT_CODE = 70


def fetch_review(review, masterbranch, remote, project):
    remote_url = get_remote_url(remote)

    review_arg = review
    review, patchset_number = parse_review_number(review)
    current_patch_set = patchset_number is None

    review_infos = query_reviews(remote_url,
                                 project=project,
                                 change=review,
                                 current_patch_set=current_patch_set,
                                 exception=CannotQueryPatchSet,
                                 parse_exc=ReviewInformationNotFound)

    if not len(review_infos):
        raise ReviewInformationNotFound(review)
    for info in review_infos:
        if 'branch' in info and info['branch'] == masterbranch:
            if VERBOSE:
                print('Using review info from branch %s' % info['branch'])
            review_info = info
            break
    else:
        review_info = review_infos[0]
        if VERBOSE and 'branch' in review_info:
            print('Using default branch %s' % review_info['branch'])

    try:
        if patchset_number is None:
            refspec = review_info['currentPatchSet']['ref']
        else:
            refspec = [ps for ps in review_info['patchSets']
                       if int(ps['number']) == int(patchset_number)][0]['ref']
    except IndexError:
        raise PatchSetNotFound(review_arg)
    except KeyError:
        raise ReviewNotFound(review)

    change_number = review_info.get('number',
                                    review_info.get('_number', review))
    remote_branch = review_info['branch']

    if patchset_number is None:
        branch_name = "review/%s" % (change_number,)
    else:
        branch_name = "review/%s-patch%s" % (change_number, patchset_number)

    print("Downloading %s from gerrit" % refspec)
    run_command_exc(PatchSetGitFetchFailed,
                    "git", "fetch", remote_url, refspec)
    return branch_name, remote_branch


def checkout_review(branch_name, remote, remote_branch):
    """Checkout a newly fetched (FETCH_HEAD) change
       into a branch
    """

    try:
        run_command_exc(CheckoutNewBranchFailed,
                        "git", "checkout", "-b",
                        branch_name, "FETCH_HEAD")
        # --set-upstream-to is supported starting in git 1.8
        if remote is not None:
            run_command_exc(SetUpstreamBranchFailed,
                            "git", "branch", "--set-upstream-to",
                            '%s/%s' % (remote, remote_branch),
                            branch_name)

    except CheckoutNewBranchFailed as e:
        if re.search(r"already exists\.?", e.output):
            print("Branch %s already exists - reusing" % branch_name)
            track_remote, track_branch = parse_tracking(
                ref='refs/heads/' + branch_name)
            if track_remote and not (track_remote == remote and
                                     track_branch == remote_branch):
                print("Branch %s incorrectly tracking %s/%s instead of %s/%s"
                      % (branch_name,
                         track_remote, track_branch,
                         remote, remote_branch))
                raise BranchTrackingMismatch
            run_command_exc(CheckoutExistingBranchFailed,
                            "git", "checkout", branch_name)
            run_command_exc(ResetKeepFailed,
                            "git", "reset", "--keep", "FETCH_HEAD")
        else:
            raise

    print("Switched to branch \"%s\"" % branch_name)


class PatchSetGitCherrypickFailed(CommandFailed):
    "There was a problem applying changeset contents to the current branch."
    EXIT_CODE = 69


def cherrypick_review(option=None):
    cmd = ["git", "cherry-pick"]
    if option:
        cmd.append(option)
    cmd.append("FETCH_HEAD")
    print(run_command_exc(PatchSetGitCherrypickFailed, *cmd))


class CheckoutBackExistingBranchFailed(CommandFailed):
    "Cannot switch back to existing branch"
    EXIT_CODE = 67


class DeleteBranchFailed(CommandFailed):
    "Failed to delete branch"
    EXIT_CODE = 68


class InvalidPatchsetsToCompare(GitReviewException):
    def __init__(self, patchsetA, patchsetB):
        Exception.__init__(
            self,
            "Invalid patchsets for comparison specified (old=%s,new=%s)" % (
                patchsetA,
                patchsetB))
    EXIT_CODE = 39


def compare_review(review_spec, branch, remote, project, rebase=False):
    new_ps = None    # none means latest

    if '-' in review_spec:
        review_spec, new_ps = review_spec.split('-')
    review, old_ps = parse_review_number(review_spec)

    if old_ps is None or old_ps == new_ps:
        raise InvalidPatchsetsToCompare(old_ps, new_ps)

    old_review = build_review_number(review, old_ps)
    new_review = build_review_number(review, new_ps)

    old_branch, _ = fetch_review(old_review, branch, remote, project)
    checkout_review(old_branch, None, None)

    if rebase:
        print('Rebasing %s' % old_branch)
        rebase = rebase_changes(branch, remote, False)
        if not rebase:
            print('Skipping rebase because of conflicts')
            run_command_exc(CommandFailed, 'git', 'rebase', '--abort')

    new_branch, remote_branch = fetch_review(
        new_review,
        branch,
        remote,
        project)
    checkout_review(new_branch, remote, remote_branch)

    if rebase:
        print('Rebasing also %s' % new_branch)
        if not rebase_changes(branch, remote, False):
            print("Rebasing of the new branch failed, "
                  "diff can be messed up (use -R to not rebase at all)!")
            run_command_exc(CommandFailed, 'git', 'rebase', '--abort')

    subprocess.check_call(['git', 'diff', old_branch])


def finish_branch(target_branch):
    local_branch = get_branch_name(target_branch)
    if VERBOSE:
        print("Switching back to '%s' and deleting '%s'" % (target_branch,
                                                            local_branch))
    run_command_exc(CheckoutBackExistingBranchFailed,
                    "git", "checkout", target_branch)
    print("Switched to branch '%s'" % target_branch)

    run_command_exc(DeleteBranchFailed,
                    "git", "branch", "-D", local_branch)
    print("Deleted branch '%s'" % local_branch)


def convert_bool(one_or_zero):
    "Return a bool on a one or zero string."
    return str(one_or_zero) in ["1", "true", "True"]


class MalformedInput(GitReviewException):
    EXIT_CODE = 3


def assert_valid_hashtags(hashtags):
    """Ensure no whitespace is found in hashtags, as it will result
    in an invalid refspec.
    """
    assert_valid_whitespace(hashtags, "hashtags")


def assert_valid_reviewers(reviewers):
    """Ensure no whitespace is found in reviewer names, as it will result
    in an invalid refspec.
    """
    assert_valid_whitespace(reviewers, "reviewers")


def assert_valid_whitespace(values, type_name):
    """Ensure no whitespace is found in list values, as it will result
    in an invalid refspec.
    """
    for v in values:
        if re.search(r"\s", v):
            raise MalformedInput(
                "Whitespace not allowed in %s: '%s'" % type_name, v)


class _DownloadFlag(argparse.Action):
    """Special action for the various forms of downloading reviews.

    Additional option parsing: store value in 'dest', but
    at the same time set one of the flag options to True
    """
    def __call__(self, parser, namespace, value, option_string=None):
        url = urlparse(value)
        # Turn URLs into change ids:
        #   https://review.opendev.org/423436
        # and
        #   https://review.opendev.org/423436/
        # and
        #   https://review.opendev.org/#/c/423436
        # and
        #   https://review.opendev.org/c/<project>/+/423436
        # become
        #   "423436"
        # while
        #   https://review.opendev.org/423436/1
        # and
        #   https://review.opendev.org/#/c/423436/1
        # and
        #   https://review.opendev.org/c/<project>/+/423436/1
        # become
        #   "423436,1".
        #
        # If there is a #, the rest of the path is stored in the
        # "fragment", otherwise that will be empty.
        base = url.fragment or url.path
        parts = base.rstrip('/').lstrip('/c').split('/')
        # PolyGerrit places the change after a '+' symbol in the url
        try:
            parts = parts[parts.index('+') + 1:]
        except ValueError:
            pass
        change = parts[0]
        if len(parts) > 1:
            change = '%s,%s' % (change, parts[1])
        setattr(namespace, self.dest, change)
        setattr(namespace, self.const, True)


def _main():
    usage = "git review [OPTIONS] ... [BRANCH]"

    description = """\
A git command for submitting branches to Gerrit.

git-review is a tool that helps submitting git branches to gerrit for
review.
"""

    epilog = """\
additional information:

  Options configurable using: git config --global gitreview.OPTION VALUE

  branchauthor          name (default), username, email
"""

    parser = argparse.ArgumentParser(
        usage=usage,
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=epilog)

    topic_arg_group = parser.add_mutually_exclusive_group()
    topic_arg_group.add_argument("-t", "--topic", dest="topic",
                                 help="Topic to submit branch to")
    topic_arg_group.add_argument("-T", "--no-topic", dest="notopic",
                                 action="store_true",
                                 help="No topic except if explicitly provided"
                                 " (deprecated)")

    parser.add_argument("--hashtags", nargs="+",
                        help="Hashtags to submit branch to")
    parser.add_argument("--reviewers", nargs="+",
                        help="Add reviewers to uploaded patchsets.")
    parser.add_argument("--cc", nargs="+",
                        help="Add CC to uploaded patchsets.")
    parser.add_argument("-n", "--dry-run", dest="dry", action="store_true",
                        help="Don't actually submit the branch for review")
    parser.add_argument("-i", "--new-changeid", dest="regenerate",
                        action="store_true",
                        help="Regenerate Change-id before submitting")
    parser.add_argument("-r", "--remote", dest="remote",
                        help="git remote to use for gerrit")
    parser.add_argument("--remote-hook", dest="remote_hook",
                        action="store_true",
                        help="Fetch the remote version of the commit-msg hook")
    parser.add_argument("--use-pushurl", dest="usepushurl",
                        action="store_true",
                        help="Use remote push-url logic instead of separate"
                             " remotes")
    parser.add_argument('--notify',
                        choices=['NONE', 'OWNER', 'OWNER_REVIEWERS', 'ALL'],
                        help="Control to whom email notifications are sent,"
                             " defaults to ALL.")
    parser.add_argument("--message", dest="message",
                        help="Message to add to patchset description")

    rebase_group = parser.add_mutually_exclusive_group()
    rebase_group.add_argument("-R", "--no-rebase", dest="rebase",
                              action="store_false",
                              help="Don't test for remote merge conflicts"
                                   " before pushing.")
    rebase_group.add_argument("-F", "--force-rebase", dest="force_rebase",
                              action="store_true",
                              help="Force and push a rebase even when not"
                                   " needed.")
    rebase_group.add_argument("-K", "--keep-rebase", dest="keep_rebase",
                              action="store_true",
                              help="Keep the unfinished test rebase if a "
                                   " merge conflict is detected.")

    track_group = parser.add_mutually_exclusive_group()
    track_group.add_argument("--track", dest="track",
                             action="store_true",
                             help="Use tracked branch as default.")
    track_group.add_argument("--no-track", dest="track",
                             action="store_false",
                             help="Ignore tracked branch.")

    fetch = parser.add_mutually_exclusive_group()
    fetch.set_defaults(download=False, compare=False, cherrypickcommit=False,
                       cherrypickindicate=False, cherrypickonly=False)
    fetch.add_argument("-d", "--download", dest="changeidentifier",
                       action=_DownloadFlag, metavar="CHANGE[,PS]",
                       const="download",
                       help="Download the contents of an existing gerrit "
                            "review into a branch. Include the patchset "
                            "number to download a specific version of the "
                            "change. The default is to take the most recent "
                            "version.")
    fetch.add_argument("-x", "--cherrypick", dest="changeidentifier",
                       action=_DownloadFlag, metavar="CHANGE",
                       const="cherrypickcommit",
                       help="Apply the contents of an existing gerrit "
                             "review onto the current branch and commit "
                             "(cherry pick; not recommended in most "
                             "situations)")
    fetch.add_argument("-X", "--cherrypickindicate", dest="changeidentifier",
                       action=_DownloadFlag, metavar="CHANGE",
                       const="cherrypickindicate",
                       help="Apply the contents of an existing gerrit "
                       "review onto the current branch and commit, "
                       "indicating its origin")
    fetch.add_argument("-N", "--cherrypickonly", dest="changeidentifier",
                       action=_DownloadFlag, metavar="CHANGE",
                       const="cherrypickonly",
                       help="Apply the contents of an existing gerrit "
                       "review to the working directory and prepare "
                       "for commit")
    fetch.add_argument("-m", "--compare", dest="changeidentifier",
                       action=_DownloadFlag, metavar="CHANGE,PS[-NEW_PS]",
                       const="compare",
                       help="Download specified and latest (or NEW_PS) "
                       "patchsets of an existing gerrit review into "
                       "a branches, rebase on master "
                       "(skipped on conflicts or when -R is specified) "
                       "and show their differences")

    parser.add_argument("-u", "--update", dest="update", action="store_true",
                        help="Force updates from remote locations")
    parser.add_argument("-s", "--setup", dest="setup", action="store_true",
                        help="Just run the repo setup commands but don't "
                             "submit anything")
    parser.add_argument("-f", "--finish", dest="finish", action="store_true",
                        help="Close down this branch and switch back to "
                             "master on successful submission")
    parser.add_argument("-l", "--list", dest="list", action="count",
                        help="List available reviews for the current project, "
                        "if passed more than once, will show more information")
    parser.add_argument("--no-thin", dest="no_thin", action="store_true",
                        help="git push with --no-thin. This may workaround "
                        "issues with pushing in some circumstances.")
    parser.add_argument("-y", "--yes", dest="yes", action="store_true",
                        help="Indicate that you do, in fact, understand if "
                             "you are submitting more than one patch")
    parser.add_argument("-v", "--verbose", dest="verbose", action="store_true",
                        help="Output more information about what's going on")

    wip_group = parser.add_mutually_exclusive_group()
    wip_group.add_argument("-w", "--wip", "--work-in-progress", dest="wip",
                           action="store_true",
                           help="Send patch as work in progress for Gerrit "
                                "versions >= 2.15")
    wip_group.add_argument("-W", "--ready", dest="ready", action="store_true",
                           help="Send patch that is already work in progress"
                                " as ready for review. Gerrit versions >="
                                " 2.15")

    private_group = parser.add_mutually_exclusive_group()
    private_group.add_argument("-p", "--private", dest="private",
                               action="store_true",
                               help="Send patch as a private patch ready for "
                                    "review. Gerrit versions >= 2.15")
    private_group.add_argument("-P", "--remove-private", dest="remove_private",
                               action="store_true",
                               help="Send patch which already in private state"
                                    " to normal patch."" Gerrit versions >= "
                                    "2.15")

    parser.add_argument("--no-custom-script", dest="custom_script",
                        action="store_false", default=True,
                        help="Do not run custom scripts.")
    parser.add_argument("--color", dest="color", metavar="<when>",
                        nargs="?", choices=["always", "never", "auto"],
                        help="Show color output. --color (without [<when>]) "
                             "is the same as --color=always. <when> can be "
                             "one of %(choices)s. Behaviour can also be "
                             "controlled by the color.ui and color.review "
                             "configuration settings.")
    parser.add_argument("--no-color", dest="color", action="store_const",
                        const="never",
                        help="Turn off colored output. Can be used to "
                             "override configuration options. Same as "
                             "setting --color=never.")
    parser.add_argument("--license", dest="license", action="store_true",
                        help="Print the license and exit")
    parser.add_argument("--version", action="version",
                        version='%s version %s' %
                        (os.path.split(sys.argv[0])[-1], get_version()))
    parser.add_argument("branch", nargs="?")

    parser.set_defaults(dry=False,
                        verbose=False,
                        update=False,
                        setup=False,
                        list=False,
                        yes=False,
                        wip=False,
                        ready=False,
                        private=False,
                        remove_private=False)

    try:
        (top_dir, git_dir) = git_directories()
    except GitDirectoriesException as _no_git_dir:
        no_git_dir = _no_git_dir
    else:
        no_git_dir = False
        config = Config(os.path.join(top_dir, ".gitreview"))
        parser.set_defaults(rebase=convert_bool(config['rebase']),
                            track=convert_bool(config['track']),
                            remote=None,
                            notopic=convert_bool(config['notopic']),
                            usepushurl=convert_bool(config['usepushurl']))
    options = parser.parse_args()

    if options.license:
        print(COPYRIGHT)
        sys.exit(0)

    if no_git_dir:
        raise no_git_dir

    if options.branch is None:
        branch = config['branch']
    else:
        # explicitly-specified branch on command line overrides options.track
        branch = options.branch
        options.track = False

    get_git_version()

    global VERBOSE
    global UPDATE
    VERBOSE = options.verbose
    UPDATE = options.update
    remote = options.remote
    if not remote:
        if options.usepushurl:
            remote = 'origin'
        else:
            remote = config['remote']
    yes = options.yes
    status = 0

    global BRANCHAUTHOR
    BRANCHAUTHOR = config['branchauthor']
    if BRANCHAUTHOR not in ["name", "email", "username"]:
        warn("Invalid gitreview.branchauthor: %s (using default: %s)"
             % (BRANCHAUTHOR, DEFAULTS["branchauthor"]))
        BRANCHAUTHOR = DEFAULTS["branchauthor"]

    if options.track:
        remote, branch = resolve_tracking(remote, branch)

    check_remote(branch, remote, config['scheme'],
                 config['hostname'], config['port'], config['project'],
                 usepushurl=options.usepushurl)

    if options.color:
        set_color_output(options.color)

    if options.changeidentifier:
        if options.compare:
            compare_review(options.changeidentifier,
                           branch, remote, config['project'],
                           options.rebase)
            return
        local_branch, remote_branch = fetch_review(options.changeidentifier,
                                                   branch, remote,
                                                   config['project'])
        if options.download:
            checkout_review(local_branch, remote, remote_branch)
        else:
            if options.cherrypickcommit:
                cherrypick_review()
            elif options.cherrypickonly:
                cherrypick_review("-n")
            if options.cherrypickindicate:
                cherrypick_review("-x")
        return
    elif options.list:
        with_topic = options.list > 1
        list_reviews(remote, config['project'],
                     branch if options.branch or options.track else None,
                     with_topic=with_topic)
        return

    if options.custom_script:
        run_custom_script("pre")

    hook_dir = git_get_hooks_path(top_dir, git_dir)
    hook_file = os.path.join(hook_dir, "commit-msg")
    have_hook = os.path.exists(hook_file) and os.access(hook_file, os.X_OK)

    if not have_hook:
        if options.remote_hook:
            set_hooks_commit_msg(remote, hook_file)
        else:
            set_hooks_commit_msg(None, hook_file)

    if options.setup:
        if options.finish and not options.dry:
            finish_branch(branch)
        return

    if options.rebase or options.force_rebase:
        if not rebase_changes(branch, remote, keep=options.keep_rebase):
            if not options.keep_rebase:
                abort_rebase()
            sys.exit(1)
        if not options.force_rebase and not undo_rebase():
            sys.exit(1)
    assert_one_change(remote, branch, yes, have_hook)

    no_thin = ''
    if options.no_thin:
        no_thin = '--no-thin'

    color_remote = ''
    if not options.dry and check_use_color_output():
        color_remote = '-c color.remote=always'

    ref = "for"

    cmd = ("git %s push --no-follow-tags %s %s HEAD:refs/%s/%s" %
           (color_remote, no_thin, remote, ref, branch))
    push_options = []
    if options.topic is not None:
        topic = options.topic
    else:
        topic = None

    if topic and topic != branch:
        push_options.append("topic=%s" % topic)

    if options.hashtags:
        assert_valid_hashtags(options.hashtags)
        push_options += ["t=%s" % r for r in options.hashtags]

    if options.reviewers:
        assert_valid_reviewers(options.reviewers)
        push_options += ["r=%s" % r for r in options.reviewers]

    if options.cc:
        assert_valid_reviewers(options.cc)
        push_options += ["cc=%s" % cc for cc in options.cc]

    if options.regenerate:
        print("Amending the commit to regenerate the change id\n")
        regenerate_cmd = "git commit --amend"
        if options.dry:
            print("\tGIT_EDITOR=\"sed -i -e '/^Change-Id:/d'\" %s\n" %
                  regenerate_cmd)
        else:
            run_command(regenerate_cmd,
                        GIT_EDITOR="sed -i -e "
                        "'/^Change-Id:/d'")

    if options.wip:
        push_options.append('wip')

    if options.ready:
        push_options.append('ready')

    if options.private:
        push_options.append('private')

    if options.remove_private:
        push_options.append('remove-private')

    if options.notify is not None:
        push_options.append("notify=%s" % options.notify)

    if options.message is not None:
        escaped_message = quote_plus(options.message)
        push_options.append(f"m={escaped_message}")

    if push_options:
        cmd += "%" + ",".join(push_options)
    if options.dry:
        print("Please use the following command "
              "to send your commits to review:\n")
        print("\t%s\n" % cmd)
    else:
        (status, output) = run_command_status(cmd)
        print(output)
        if status == 1 and 'Missing tree' in output and not options.no_thin:
            print('Consider trying again with --no-thin')

    if options.finish and not options.dry and status == 0:
        finish_branch(branch)
        return

    if options.custom_script:
        run_custom_script("post")
    sys.exit(status)


def main():

    try:
        _main()
    except GitReviewException as e:
        print(e)
        sys.exit(e.EXIT_CODE)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import shutil
import stat
import tempfile

from git_review.hooks import COMMIT_MSG
from git_review import tests
from git_review.tests import utils


class ConfigTestCase(tests.BaseGitReviewTestCase):
    """Class for config tests."""

    def test_get_config_from_cli(self):
        self.reset_remote()
        self._run_git('remote', 'rm', 'origin')
        self._create_gitreview_file(defaultremote='remote-file')
        self._run_git('config', 'gitreview.remote', 'remote-gitconfig')
        self._run_git_review('-s', '-r', 'remote-cli')

        remote = self._run_git('remote').strip()
        self.assertEqual('remote-cli', remote)

    def test_get_config_from_gitconfig(self):
        self.reset_remote()
        self._run_git('remote', 'rm', 'origin')
        self._create_gitreview_file(defaultremote='remote-file')
        self._run_git('config', 'gitreview.remote', 'remote-gitconfig')
        self._run_git_review('-s')

        remote = self._run_git('remote').strip()
        self.assertEqual('remote-gitconfig', remote)

    def test_get_config_from_file(self):
        self.reset_remote()
        self._run_git('remote', 'rm', 'origin')
        self._create_gitreview_file(defaultremote='remote-file')
        self._run_git_review('-s')

        remote = self._run_git('remote').strip()
        self.assertEqual('remote-file', remote)


class GitReviewTestCase(tests.BaseGitReviewTestCase):
    """Class for the git-review tests."""

    def test_cloned_repo(self):
        """Test git-review on the just cloned repository."""
        self._simple_change('test file modified', 'test commit message')
        self.assertNotIn('Change-Id:', self._run_git('log', '-1'))
        gr = self._run_git_review()
        self.assertIn('remote: SUCCESS', gr)
        self.assertIn('test commit message [NEW]', gr)
        self.assertIn('Change-Id:', self._run_git('log', '-1'))

    def test_git_review_version(self):
        """Test git-review --version."""
        self.assertTrue(
            self._run_git_review('--version').startswith('git-review version'))

    def test_git_review_s(self):
        """Test git-review -s."""
        self.reset_remote()
        self._run_git_review('-s')
        self._simple_change('test file modified', 'test commit message')
        self.assertIn('Change-Id:', self._run_git('log', '-1'))

    def test_git_review_s_in_detached_head(self):
        """Test git-review -s in detached HEAD state."""
        self.reset_remote()
        master_sha1 = self._run_git('rev-parse', 'master')
        self._run_git('checkout', master_sha1)
        self._run_git_review('-s')
        self._simple_change('test file modified', 'test commit message')
        self.assertIn('Change-Id:', self._run_git('log', '-1'))

    def test_git_review_s_with_outdated_repo(self):
        """Test git-review -s with a outdated repo."""
        # Push simple change to gerrit and submit it
        hook_file = self._dir('test', '.git/hooks/commit-msg')
        utils.write_to_file(hook_file, COMMIT_MSG.encode())
        os.chmod(hook_file, os.stat(hook_file).st_mode | stat.S_IEXEC)
        self._simple_change('test file to outdate', 'test commit message 1')
        self._run_git('push', 'origin', 'HEAD:refs/for/master')
        # Cleanup the hook file so that setup below rewrites it
        os.unlink(hook_file)
        # Submit the proposed change
        head = self._run_git('rev-parse', 'HEAD')
        self._run_gerrit_cli('review', head, '--code-review=+2', '--submit')
        # Reset so that the local repo is behind the upstream repo in Gerrit.
        self._run_git('reset', '--hard', 'HEAD^')

        # Review setup with an outdated repo
        self.reset_remote()
        self._run_git_review('-s')
        self._simple_change('test file modified', 'test commit message 2')
        self.assertIn('Change-Id:', self._run_git('log', '-1'))

    def test_git_review_s_from_subdirectory(self):
        """Test git-review -s from subdirectory."""
        self.reset_remote()
        utils.run_cmd('mkdir', 'subdirectory', chdir=self.test_dir)
        self._run_git_review(
            '-s', chdir=os.path.join(self.test_dir, 'subdirectory'))

    def test_install_embedded_hook(self):
        """Test whether git-review -s correctly creates the commit-msg hook
        from the embedded copy with appropriate permissions and content.
        """
        hooks_subdir = ".git/hooks"
        hook_file = os.path.join(self.test_dir, hooks_subdir, 'commit-msg')
        self.reset_remote()

        self._run_git_review('-s')
        self.assertTrue(os.path.exists(hook_file))
        self.assertTrue(os.access(hook_file, os.W_OK))
        self.assertTrue(os.access(hook_file, os.X_OK))
        with open(hook_file) as f:
            content = f.read()
            self.assertTrue(content.startswith('#!/'))
            # This line should not appear in the embedded version
            self.assertNotIn('# From Gerrit Code Review', content)

    def test_install_remote_hook(self):
        """Test whether git-review -s correctly creates the commit-msg hook
        from the Gerrit server with appropriate permissions and content.
        """
        hooks_subdir = ".git/hooks"
        hook_file = os.path.join(self.test_dir, hooks_subdir, 'commit-msg')
        self.reset_remote()

        self._run_git_review('-s', '--remote-hook')
        self.assertTrue(os.path.exists(hook_file))
        self.assertTrue(os.access(hook_file, os.W_OK))
        self.assertTrue(os.access(hook_file, os.X_OK))
        with open(hook_file) as f:
            content = f.read()
            self.assertTrue(content.startswith('#!/'))
            # This line is injected by the Gerrit server in its version
            self.assertIn('# From Gerrit Code Review', content)

    def test_git_review_s_without_core_hooks_path_option(self):
        """Test whether git-review -s correctly creates the commit-msg hook,
        with the Git core.hooksPath option unset.
        """
        hooks_subdir = ".git/hooks"
        self.reset_remote()

        # There really isn't a good way to ensure that
        # "core.hooksPath" option is presently unset. "git config
        # --unset" is not idempotent; if you try to unset a config
        # option that isn't defined it fails with an exit code of
        # 5. Running "git config core.hooksPath" to retrieve the value
        # returns 1 if unset, but we can't use self.assertRaises here
        # either because run_cmd raises a generic Exception and that's
        # too broad. So instead, we don't check core.hooksPath at all
        # here, and instead rely on the next two tests to unset the
        # option after they've set it.
        self._run_git_review('-s')
        self.assertTrue(os.path.exists(os.path.join(self.test_dir,
                                                    hooks_subdir,
                                                    'commit-msg')))

    def test_git_review_s_with_core_hooks_path_option_relative(self):
        """Test whether git-review -s correctly creates the commit-msg hook,
        with the Git core.hooksPath option set to a relative path.
        """
        hooks_subdir = "foo"
        self.reset_remote()
        self._run_git("config", "core.hooksPath", hooks_subdir)
        self._run_git_review('-s')
        self.assertTrue(os.path.exists(os.path.join(self.test_dir,
                                                    hooks_subdir,
                                                    'commit-msg')))
        self._run_git("config", "--unset", "core.hooksPath")

    def test_git_review_s_with_core_hooks_path_option_absolute(self):
        """Test whether git-review -s correctly creates the commit-msg hook,
        with the Git core.hooksPath option set to an absolute path.
        """
        self.reset_remote()
        with tempfile.TemporaryDirectory() as hooks_dir:
            self._run_git("config", "core.hooksPath", hooks_dir)
            self._run_git_review('-s')
            self.assertTrue(os.path.exists(os.path.join(hooks_dir,
                                                        'commit-msg')))
        self._run_git("config", "--unset", "core.hooksPath")

    def test_git_review_d(self):
        """Test git-review -d."""
        self._run_git_review('-s')

        # create new review to be downloaded
        self._simple_change('test file modified', 'test commit message')
        self._run_git_review()
        change_id = self._run_git('log', '-1').split()[-1]

        shutil.rmtree(self.test_dir)

        # download clean Git repository and fresh change from Gerrit to it
        self._run_git('clone', self.project_uri)
        self.configure_gerrit_remote()
        self._run_git_review('-d', change_id)
        self.assertIn('test commit message', self._run_git('log', '-1'))

        # test backport branch
        self._run_git('checkout', '-b', 'mybackport',
                      self._remote + '/' + 'testbranch')
        self._simple_change('test file modified in branch',
                            'test branch commit message\n\nChange-Id: %s' %
                            change_id)
        self._run_git_review('testbranch')
        self._run_git('checkout', 'master')
        self._run_git_review('-d', change_id, 'testbranch')
        self.assertIn('test branch commit message',
                      self._run_git('log', '-1'))

        # second download should also work correctly
        self._run_git('checkout', 'master')
        self._run_git_review('-d', change_id)
        self.assertIn('test commit message', self._run_git('show', 'HEAD'))
        self.assertNotIn('test commit message',
                         self._run_git('show', 'HEAD^1'))

        # and branch is tracking
        head = self._run_git('symbolic-ref', '-q', 'HEAD')
        self.assertIn(
            'refs/remotes/%s/master' % self._remote,
            self._run_git("for-each-ref", "--format='%(upstream)'", head))

        # add some more changes & upload
        self._simple_amend('test 2nd rev',
                           self._dir('test', '2nd_rev_file.txt'))
        self._run_git_review('-v')
        self._simple_amend('test 3rd rev',
                           self._dir('test', '3rd_rev_file.txt'))
        self._run_git_review('-v')

        # get rev 2; assert rev2 file is there, but not rev3
        self._run_git_review('-d', '%s,%s' % (change_id, 2))
        self.assertIn('2nd_rev_file.txt',
                      self._run_git('show', 'HEAD'))
        self.assertNotIn('3rd_rev_file.txt',
                         self._run_git('show', 'HEAD'))

    def test_multiple_changes(self):
        """Test git-review asks about multiple changes.

        Should register user's wish to send two change requests by interactive
        'yes' message and by the -y option.
        """
        self._run_git_review('-s')

        # 'yes' message
        self._simple_change('test file modified 1st time',
                            'test commit message 1')
        self._simple_change('test file modified 2nd time',
                            'test commit message 2')

        review_res = self._run_git_review(confirm=True)
        self.assertIn("Type 'yes' to confirm", review_res)
        self.assertIn("new: 2", review_res)

        # abandon changes sent to the Gerrit
        head = self._run_git('rev-parse', 'HEAD')
        head_1 = self._run_git('rev-parse', 'HEAD^1')
        self._run_gerrit_cli('review', '--abandon', head)
        self._run_gerrit_cli('review', '--abandon', head_1)

        # -y option
        self._simple_change('test file modified 3rd time',
                            'test commit message 3')
        self._simple_change('test file modified 4th time',
                            'test commit message 4')
        review_res = self._run_git_review('-y')
        self.assertIn("new: 2", review_res)
        self.assertIn("remote: SUCCESS", review_res)
        self.assertIn("test commit message 3 [NEW]", review_res)
        self.assertIn("test commit message 4 [NEW]", review_res)

    def test_git_review_re(self):
        """Test git-review adding reviewers to changes."""
        self._run_git_review('-s')

        # Create users to add as reviewers
        self._run_gerrit_cli('create-account', '--email',
                             'reviewer1@example.com', 'reviewer1')
        self._run_gerrit_cli('create-account', '--email',
                             'reviewer2@example.com', 'reviewer2')

        self._simple_change('test file', 'test commit message')

        review_res = self._run_git_review('--reviewers', 'reviewer1',
                                          'reviewer2')
        self.assertIn("new: 1", review_res)

        # verify both reviewers are on patchset
        head = self._run_git('rev-parse', 'HEAD')
        change = self._run_gerrit_cli('query', '--format=JSON',
                                      '--all-reviewers', head)
        # The first result should be the one we want
        change = json.loads(change.split('\n')[0])

        self.assertEqual(2, len(change['allReviewers']))

        reviewers = set()
        for reviewer in change['allReviewers']:
            reviewers.add(reviewer['username'])

        self.assertEqual(set(['reviewer1', 'reviewer2']), reviewers)

    def test_git_review_cc(self):
        """Test git-review adding CC to changes."""
        self._run_git_review('-s')

        # Create users to add as CC
        self._run_gerrit_cli('create-account', '--email',
                             'cc1@example.com', 'cc1')
        self._run_gerrit_cli('create-account', '--email',
                             'cc2@example.com', 'cc2')

        self._simple_change('test file', 'test commit message')

        review_ccs = self._run_git_review('--cc', 'cc1', 'cc2')
        self.assertIn("new: 1", review_ccs)

        # verify both CCs are on patchset
        head = self._run_git('rev-parse', 'HEAD')
        change = self._run_gerrit_cli('query', '--format=JSON',
                                      '--all-reviewers', head)
        # The first result should be the one we want
        change = json.loads(change.split('\n')[0])

        self.assertEqual(2, len(change['allReviewers']))

        ccs = set()
        for cc in change['allReviewers']:
            ccs.add(cc['username'])

        self.assertEqual(set(['cc1', 'cc2']), ccs)

    def test_rebase_unstaged_changes_msg(self):
        """Test message displayed when unstaged changes are present."""
        self._run_git_review('-s')
        self._run_git('checkout', '-b', 'test_branch')
        self._unstaged_change(change_text='simple message')
        exc = self.assertRaises(Exception, self._run_git_review)
        self.assertIn("You have unstaged changes. Please", exc.args[0])

    def test_rebase_uncommitted_changes_msg(self):
        """Test message displayed when staged changes are present."""
        self._run_git_review('-s')
        self._run_git('checkout', '-b', 'test_branch')
        self._uncommitted_change(change_text='simple message')
        exc = self.assertRaises(Exception, self._run_git_review)
        self.assertIn("You have uncommitted changes. Please", exc.args[0])

    def test_ignore_unstaged_submodule_changes(self):
        """Test message displayed when unstaged changes are present."""
        self._run_git_review('-s')
        self._run_git('checkout', '-b', 'test_branch')
        self._run_git_sub('init')
        self._unstaged_change_sub(change_text='simple message')
        self._run_git_review()

    def test_ignore_uncommitted_submodule_changes(self):
        """Test message displayed when staged changes are present."""
        self._run_git_review('-s')
        self._run_git('checkout', '-b', 'test_branch')
        self._run_git_sub('init')
        self._uncommitted_change_sub(change_text='simple message')
        self._run_git_review()

    def test_rebase_no_remote_branch_msg(self):
        """Test message displayed where no remote branch exists."""
        self._run_git_review('-s')
        self._run_git('checkout', '-b', 'new_branch')
        self._simple_change('simple message',
                            'need to avoid noop message')
        exc = self.assertRaises(Exception, self._run_git_review, 'new_branch')
        self.assertIn("The branch 'new_branch' does not exist on the given "
                      "remote '%s'" % self._remote, exc.args[0])

    def test_need_rebase_no_upload(self):
        """Test change needing a rebase does not upload."""
        self._run_git_review('-s')
        head_1 = self._run_git('rev-parse', 'HEAD^1')

        self._run_git('checkout', '-b', 'test_branch', head_1)

        self._simple_change('some other message',
                            'create conflict with master')

        exc = self.assertRaises(Exception, self._run_git_review)
        rebased = ("Errors running git rebase --rebase-merges "
                   "-i remotes/%s/master" % self._remote in exc.args[0] or
                   "Errors running git rebase --preserve-merges "
                   "-i remotes/%s/master" % self._remote in exc.args[0])
        self.assertTrue(rebased)
        self.assertIn("It is likely that your change has a merge conflict",
                      exc.args[0])
        exc = self.assertRaises(Exception, self._run_git, 'rebase', '--abort')
        self.assertIn("No rebase in progress?", exc.args[0])

    def test_keep_aborted_rebase_state(self):
        """Test git-review -K behavior when rebase is needed."""
        self._run_git_review('-s')
        head_1 = self._run_git('rev-parse', 'HEAD^1')

        self._run_git('checkout', '-b', 'test_branch', head_1)

        self._simple_change('some other message',
                            'create conflict with master')

        self.assertRaises(Exception, self._run_git_review, '-K')
        self.assertIn(
            "interactive rebase in progress", self._run_git('status'))

    def test_upload_without_rebase(self):
        """Test change not needing a rebase can upload without rebasing."""
        self._run_git_review('-s')
        head_1 = self._run_git('rev-parse', 'HEAD^1')

        self._run_git('checkout', '-b', 'test_branch', head_1)

        self._simple_change('some new message',
                            'just another file (no conflict)',
                            self._dir('test', 'new_test_file.txt'))

        review_res = self._run_git_review('-v')
        rebased = ("Running: git rebase --rebase-merges "
                   "-i remotes/%s/master" % self._remote in review_res or
                   "Running: git rebase --preserve-merges "
                   "-i remotes/%s/master" % self._remote in review_res)
        self.assertTrue(rebased)
        self.assertEqual(self._run_git('rev-parse', 'HEAD^1'), head_1)

    def test_uploads_with_nondefault_rebase(self):
        """Test changes rebase against correct branches."""
        # prepare maintenance branch that is behind master
        hook_file = self._dir('test', '.git/hooks/commit-msg')
        utils.write_to_file(hook_file, COMMIT_MSG.encode())
        os.chmod(hook_file, os.stat(hook_file).st_mode | stat.S_IEXEC)
        self._create_gitreview_file(track='true',
                                    defaultremote='origin')
        self._run_git('add', '.gitreview')
        self._run_git('commit', '-m', 'track=true.')
        self._simple_change('diverge master from maint',
                            'no conflict',
                            self._dir('test', 'test_file_to_diverge.txt'))
        self._run_git('push', 'origin', 'HEAD:refs/for/master')
        os.unlink(hook_file)
        head_1 = self._run_git('rev-parse', 'HEAD^1')
        head = self._run_git('rev-parse', 'HEAD')
        self._run_gerrit_cli('review', head_1, '--code-review=+2', '--submit')
        self._run_gerrit_cli('review', head, '--code-review=+2', '--submit')
        self._run_gerrit_cli('create-branch',
                             'test/test_project',
                             'other', head)
        self._run_git_review('-s')
        head_1 = self._run_git('rev-parse', 'HEAD^1')
        self._run_gerrit_cli('create-branch',
                             'test/test_project',
                             'maint', head_1)
        self._run_git('fetch')

        self._run_git('checkout', '-b', 'test_branch', 'origin/maint')
        branches = self._run_git('branch', '-a')
        expected_branch = '* test_branch'
        observed = branches.split('\n')
        self.assertIn(expected_branch, observed)

        self._simple_change('some new message',
                            'just another file (no conflict)',
                            self._dir('test', 'new_tracked_test_file.txt'))
        change_id = self._run_git('log', '-1').split()[-1]

        review_res = self._run_git_review('-v')
        # no rebase needed; if it breaks it would try to rebase to master
        self.assertNotIn("Running: git rebase --rebase-merges "
                         "-i remotes/origin/master", review_res)
        self.assertNotIn("Running: git rebase --preserve-merges "
                         "-i remotes/origin/master", review_res)
        # Don't need to query gerrit for the branch as the second half
        # of this test will work only if the branch was correctly
        # stored in gerrit

        # delete branch locally
        self._run_git('checkout', 'master')
        self._run_git('branch', '-D', 'test_branch')

        # download, amend, submit
        self._run_git_review('-d', change_id)
        self._simple_amend('just another file (no conflict)',
                           self._dir('test', 'new_tracked_test_file_2.txt'))
        new_change_id = self._run_git('log', '-1').split()[-1]
        self.assertEqual(change_id, new_change_id)
        review_res = self._run_git_review('-v')
        # caused the right thing to happen
        rebased = ("Running: git rebase --rebase-merges "
                   "-i remotes/origin/maint" in review_res or
                   "Running: git rebase --preserve-merges "
                   "-i remotes/origin/maint" in review_res)
        self.assertTrue(rebased)

        # track different branch than expected in changeset
        branch = self._run_git('rev-parse', '--abbrev-ref', 'HEAD')
        self._run_git('branch',
                      '--set-upstream-to',
                      'remotes/origin/other',
                      branch)
        self.assertRaises(
            Exception,  # cmd.BranchTrackingMismatch inside
            self._run_git_review, '-d', change_id)

    def test_no_rebase_check(self):
        """Test -R causes a change to be uploaded without rebase checking."""
        self._run_git_review('-s')
        head_1 = self._run_git('rev-parse', 'HEAD^1')

        self._run_git('checkout', '-b', 'test_branch', head_1)
        self._simple_change('some new message', 'just another file',
                            self._dir('test', 'new_test_file.txt'))

        review_res = self._run_git_review('-v', '-R')
        self.assertNotIn('rebase', review_res)
        self.assertEqual(self._run_git('rev-parse', 'HEAD^1'), head_1)

    def test_rebase_anyway(self):
        """Test -F causes a change to be rebased regardless."""
        self._run_git_review('-s')
        head = self._run_git('rev-parse', 'HEAD')
        head_1 = self._run_git('rev-parse', 'HEAD^1')

        self._run_git('checkout', '-b', 'test_branch', head_1)
        self._simple_change('some new message', 'just another file',
                            self._dir('test', 'new_test_file.txt'))
        review_res = self._run_git_review('-v', '-F')
        self.assertIn('rebase', review_res)
        self.assertEqual(self._run_git('rev-parse', 'HEAD^1'), head)

    def _assert_branch_would_be(self, branch, extra_args=None):
        extra_args = extra_args or []
        output = self._run_git_review('-n', *extra_args)
        # last non-empty line should be:
        #       git push gerrit HEAD:refs/publish/master
        last_line = output.strip().split('\n')[-1]
        branch_was = last_line.rsplit(' ', 1)[-1].split('/', 2)[-1]
        self.assertEqual(branch, branch_was)

    def test_detached_head(self):
        """Test on a detached state: we shouldn't have '(detached' as topic."""
        self._run_git_review('-s')
        curr_branch = self._run_git('rev-parse', '--abbrev-ref', 'HEAD')
        # Note: git checkout --detach has been introduced in git 1.7.5 (2011)
        self._run_git('checkout', curr_branch + '^0')
        self._simple_change('some new message', 'just another file',
                            self._dir('test', 'new_test_file.txt'))
        # switch to French, 'git branch' should return '(détaché du HEAD)'
        lang_env = os.getenv('LANG', 'C')
        os.environ.update(LANG='fr_FR.UTF-8')
        try:
            self._assert_branch_would_be(curr_branch)
        finally:
            os.environ.update(LANG=lang_env)

    def test_no_topic(self):
        """Test on change with no topic.

        This will be checked out as 'review/{owner}/{ID}'. We don't want to set
        the topic to '{ID}'.
        """
        self._run_git_review('-s')
        curr_branch = self._run_git('rev-parse', '--abbrev-ref', 'HEAD')
        self._run_git('checkout', '-b', 'review/johndoe/123456')
        self._simple_change('test file modified', 'derp derp derp')
        self._assert_branch_would_be(curr_branch)

    def test_git_review_t(self):
        self._run_git_review('-s')
        self._simple_change('test file modified', 'commit message for bug 654')
        self._assert_branch_would_be('master%topic=zat',
                                     extra_args=['-t', 'zat'])

        # -t takes precedence over notopic
        self._run_git('config', 'gitreview.notopic', 'true')
        self._assert_branch_would_be('master%topic=zat',
                                     extra_args=['-t', 'zat'])

    def test_git_review_T(self):
        # This option is deprecated and kept for compatibility; the
        # only behavior now is "notopic".
        self._run_git_review('-s')
        self._run_git('checkout', '-b', 'bug/456')
        self._simple_change('test file modified', 'commit message for bug 456')
        self._assert_branch_would_be('master')
        self._assert_branch_would_be('master', extra_args=['-T'])

        self._run_git('config', 'gitreview.notopic', 'true')
        self._assert_branch_would_be('master')
        self._run_git('config', 'gitreview.notopic', 'false')
        self._assert_branch_would_be('master')

        self._assert_branch_would_be('master', extra_args=['-T'])

    def test_git_review_T_t(self):
        self.assertRaises(Exception, self._run_git_review, '-T', '-t', 'taz')

    def test_git_review_l(self):
        self._run_git_review('-s')

        # Populate "project" repo
        self._simple_change('project: test1', 'project: change1, merged')
        self._simple_change('project: test2', 'project: change2, open')
        self._simple_change('project: test3', 'project: change3, abandoned')
        self._run_git_review('-y')
        head = self._run_git('rev-parse', 'HEAD')
        head_2 = self._run_git('rev-parse', 'HEAD^^')
        self._run_gerrit_cli('review', head_2, '--code-review=+2', '--submit')
        self._run_gerrit_cli('review', head, '--abandon')

        # Populate "project2" repo
        self._run_gerrit_cli('create-project', 'test/test_project2',
                             '--empty-commit')
        project2_uri = self.project_uri.replace('test/test_project',
                                                'test/test_project2')
        self._run_git('fetch', project2_uri, 'HEAD')
        self._run_git('checkout', 'FETCH_HEAD')
        # We have to rewrite the .gitreview file after this checkout.
        self._create_gitreview_file()
        self._simple_change('project2: test1', 'project2: change1, open')
        self._run_git('push', project2_uri, 'HEAD:refs/for/master')

        # Only project1 open changes
        result = self._run_git_review('-l')
        self.assertNotIn('project: change1, merged', result)
        self.assertIn('project: change2, open', result)
        self.assertNotIn('project: change3, abandoned', result)
        self.assertNotIn('project2:', result)

    def _test_git_review_F(self, rebase):
        self._run_git_review('-s')

        # Populate repo
        self._simple_change('create file', 'test commit message')
        change1 = self._run_git('rev-parse', 'HEAD')
        self._run_git_review()
        self._run_gerrit_cli('review', change1, '--code-review=+2', '--submit')
        self._run_git('reset', '--hard', 'HEAD^')

        # Review with force_rebase
        self._run_git('config', 'gitreview.rebase', rebase)
        self._simple_change('create file2', 'test commit message 2',
                            self._dir('test', 'test_file2.txt'))
        self._run_git_review('-F')
        head_1 = self._run_git('rev-parse', 'HEAD^')
        self.assertEqual(change1, head_1)

    def test_git_review_F(self):
        self._test_git_review_F('1')

    def test_git_review_F_norebase(self):
        self._test_git_review_F('0')

    def test_git_review_F_R(self):
        self.assertRaises(Exception, self._run_git_review, '-F', '-R')

    def test_git_review_no_thin(self):
        """Test git-review --no-thin."""
        self._run_git_review('-s')

        # Push with --no-thin set. We can't really introspect the packs
        # that were sent so we infer success as the command not failing.
        self._simple_change('test file modified', 'test commit message')
        self._run_git_review('--no-thin')

    def test_config_instead_of_honored(self):
        self.set_remote('test_project_url')

        self.assertRaises(Exception, self._run_git_review, '-l')

        self._run_git('config', '--add', 'url.%s.insteadof' % self.project_uri,
                      'test_project_url')
        self._run_git_review('-l')

    def test_config_pushinsteadof_honored(self):
        self.set_remote('test_project_url')

        self.assertRaises(Exception, self._run_git_review, '-l')

        self._run_git('config', '--add',
                      'url.%s.pushinsteadof' % self.project_uri,
                      'test_project_url')
        self._run_git_review('-l')


class PushUrlTestCase(GitReviewTestCase):
    """Class for the git-review tests using origin push-url."""

    _remote = 'origin'

    def set_remote(self, uri):
        self._run_git('remote', 'set-url', '--push', self._remote, uri)

    def reset_remote(self):
        self._run_git('config', '--unset', 'remote.%s.pushurl' % self._remote)

    def configure_gerrit_remote(self):
        self.set_remote(self.project_uri)
        self._run_git('config', 'gitreview.usepushurl', '1')

    def test_config_pushinsteadof_honored(self):
        self.skipTest("pushinsteadof doesn't rewrite pushurls")


class HttpGitReviewTestCase(tests.HttpMixin, GitReviewTestCase):
    """Class for the git-review tests over HTTP(S)."""
    pass



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import sys

from git_review import tests
from git_review.tests import utils


def list_test_ids(argv):
    res = utils.run_cmd(sys.executable, '-m', 'testtools.run', *argv[1:])
    return res.split('\n')


def find_collisions(test_ids):
    hashes = {}
    for test_id in test_ids:
        hash_ = tests._hash_test_id(test_id)
        if hash_ in hashes:
            return (hashes[hash_], test_id)
        hashes[hash_] = test_id
    return None


def main(argv):
    test_ids = list_test_ids(argv)
    if not test_ids:
        print("No tests found, check command line arguments", file=sys.stderr)
        return 1
    collision = find_collisions(test_ids)
    if collision is None:
        return 0
    print(
        "Found a collision for test ids hash function: %s and %s\n"
        "You should change _hash_test_id function in"
        " git_review/tests/__init__.py module to fit new set of test ids."
        % collision,
        file=sys.stderr,
    )
    return 2


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# -*- coding: utf-8 -*-

# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import functools
import io
import os
import sys
import textwrap
from unittest import mock

import fixtures
import testtools

from git_review import cmd
from git_review.tests import IsoEnvDir
from git_review.tests import utils


class ConfigTestCase(testtools.TestCase):
    """Class testing config behavior."""

    @mock.patch('git_review.cmd.LOCAL_MODE',
                mock.PropertyMock(return_value=True))
    @mock.patch('git_review.cmd.git_directories', return_value=['', 'fake'])
    @mock.patch('git_review.cmd.run_command_exc')
    def test_git_local_mode(self, run_mock, dir_mock):
        cmd.git_config_get_value('abc', 'def')
        run_mock.assert_called_once_with(
            cmd.GitConfigException,
            'git', 'config', '-f', 'fake/config', '--get', 'abc.def')

    @mock.patch('git_review.cmd.LOCAL_MODE',
                mock.PropertyMock(return_value=True))
    @mock.patch('os.path.exists', return_value=False)
    def test_gitreview_local_mode(self, exists_mock):
        cmd.Config()
        self.assertFalse(exists_mock.called)


class GitReviewConsole(testtools.TestCase, fixtures.TestWithFixtures):
    """Class for testing the console output of git-review."""

    reviews = [
        {
            'number': '1010101',
            'branch': 'master',
            'subject': 'A simple short subject',
            'topic': 'simple-topic'
        }, {
            'number': 9877,  # Starting with 2.14, numbers are sent as int
            'branch': 'stable/codeword',
            'subject': 'A longer and slightly more wordy subject'
        }, {
            'number': '12345',
            'branch': 'master',
            'subject': 'A ridiculously long subject that can exceed the '
                       'normal console width, just need to ensure the '
                       'max width is short enough'
        }]

    def setUp(self):
        # will set up isolated env dir
        super(GitReviewConsole, self).setUp()

        # ensure all tests get a separate git dir to work in to avoid
        # local git config from interfering
        iso_env = self.useFixture(IsoEnvDir())

        self._run_git = functools.partial(utils.run_git,
                                          chdir=iso_env.work_dir)

        self.run_cmd_patcher = mock.patch('git_review.cmd.run_command_status')
        run_cmd_partial = functools.partial(
            cmd.run_command_status, GIT_WORK_TREE=iso_env.work_dir,
            GIT_DIR=os.path.join(iso_env.work_dir, '.git'))
        self.run_cmd_mock = self.run_cmd_patcher.start()
        self.run_cmd_mock.side_effect = run_cmd_partial

        self._run_git('init')
        self._run_git('commit', '--allow-empty', '-m "initial commit"')
        self._run_git('commit', '--allow-empty', '-m "2nd commit"')

    def tearDown(self):
        self.run_cmd_patcher.stop()
        super(GitReviewConsole, self).tearDown()

    @mock.patch('git_review.cmd.get_version',
                side_effect=cmd.GitReviewException(u"simple-toπ㌀c"))
    def test_print_exception_with_unicode(self, exists_mock):

        try:
            with mock.patch('sys.argv', ['git-review', '--version']):
                with self.assertRaisesRegex(SystemExit, '1'):
                    cmd.main()
        except Exception as e:
            self.fail('Exception not expected: %s' % e)

    @mock.patch('git_review.cmd.query_reviews')
    @mock.patch('git_review.cmd.get_remote_url', mock.MagicMock)
    @mock.patch('git_review.cmd._has_color', False)
    def test_list_reviews_output(self, mock_query):

        mock_query.return_value = self.reviews
        with mock.patch('sys.stdout', new_callable=io.StringIO) as output:
            cmd.list_reviews(None, None)
            console_output = output.getvalue().split('\n')

        self.assertEqual(
            ['1010101           master  A simple short subject',
             '   9877  stable/codeword  A longer and slightly more wordy '
             'subject'],
            console_output[:2])

    @mock.patch('git_review.cmd.query_reviews')
    @mock.patch('git_review.cmd.get_remote_url', mock.MagicMock)
    @mock.patch('git_review.cmd._has_color', False)
    def test_list_reviews_output_with_topic(self, mock_query):

        mock_query.return_value = self.reviews
        with mock.patch('sys.stdout', new_callable=io.StringIO) as output:
            cmd.list_reviews(None, None, with_topic=True)
            console_output = output.getvalue().split('\n')

        self.assertEqual(
            ['1010101           master  simple-topic  A simple short subject',
             '   9877  stable/codeword             -  A longer and slightly '
             'more wordy subject'],
            console_output[:2])

    @mock.patch('git_review.cmd.query_reviews')
    @mock.patch('git_review.cmd.get_remote_url', mock.MagicMock)
    @mock.patch('git_review.cmd._has_color', False)
    def test_list_reviews_no_blanks(self, mock_query):

        mock_query.return_value = self.reviews
        with mock.patch('sys.stdout', new_callable=io.StringIO) as output:
            cmd.list_reviews(None, None)
            console_output = output.getvalue().split('\n')

        wrapper = textwrap.TextWrapper(replace_whitespace=False,
                                       drop_whitespace=False)
        for text in console_output:
            for line in wrapper.wrap(text):
                self.assertEqual(line.isspace(), False,
                                 "Extra blank lines appearing between reviews"
                                 "in console output")

    @mock.patch('git_review.cmd._use_color', None)
    def test_color_output_disabled(self):
        """Test disabling of colour output color.ui defaults to enabled
        """

        # git versions < 1.8.4 default to 'color.ui' being false
        # so must be set to auto to correctly test
        self._run_git("config", "color.ui", "auto")

        self._run_git("config", "color.review", "never")
        self.assertFalse(cmd.check_use_color_output(),
                         "Failed to detect color output disabled")

    @mock.patch('git_review.cmd._use_color', None)
    def test_color_output_forced(self):
        """Test force enable of colour output when color.ui
        is defaulted to false
        """

        self._run_git("config", "color.ui", "never")

        self._run_git("config", "color.review", "always")
        self.assertTrue(cmd.check_use_color_output(),
                        "Failed to detect color output forcefully "
                        "enabled")

    @mock.patch('git_review.cmd._use_color', None)
    def test_color_output_fallback(self):
        """Test fallback to using color.ui when color.review is not
        set
        """

        self._run_git("config", "color.ui", "always")
        self.assertTrue(cmd.check_use_color_output(),
                        "Failed to use fallback to color.ui when "
                        "color.review not present")


class FakeResponse(object):

    def __init__(self, code, text=""):
        self.status_code = code
        self.text = text


class FakeException(Exception):

    def __init__(self, code, *args, **kwargs):
        super(FakeException, self).__init__(*args, **kwargs)
        self.code = code


FAKE_GIT_CREDENTIAL_FILL = """\
protocol=http
host=gerrit.example.com
username=user
password=pass
"""


class ResolveTrackingUnitTest(testtools.TestCase):
    """Class for testing resolve_tracking."""
    def setUp(self):
        testtools.TestCase.setUp(self)
        patcher = mock.patch('git_review.cmd.run_command_exc')
        self.addCleanup(patcher.stop)
        self.run_command_exc = patcher.start()

    def test_track_local_branch(self):
        'Test that local tracked branch is not followed.'
        self.run_command_exc.side_effect = [
            '',
            'refs/heads/other/branch',
        ]
        self.assertEqual(cmd.resolve_tracking(u'remote', u'rbranch'),
                         (u'remote', u'rbranch'))

    def test_track_untracked_branch(self):
        'Test that local untracked branch is not followed.'
        self.run_command_exc.side_effect = [
            '',
            '',
        ]
        self.assertEqual(cmd.resolve_tracking(u'remote', u'rbranch'),
                         (u'remote', u'rbranch'))

    def test_track_remote_branch(self):
        'Test that remote tracked branch is followed.'
        self.run_command_exc.side_effect = [
            '',
            'refs/remotes/other/branch',
        ]
        self.assertEqual(cmd.resolve_tracking(u'remote', u'rbranch'),
                         (u'other', u'branch'))

    def test_track_git_error(self):
        'Test that local tracked branch is not followed.'
        self.run_command_exc.side_effect = [cmd.CommandFailed(1, '', [], {})]
        self.assertRaises(cmd.CommandFailed,
                          cmd.resolve_tracking, u'remote', u'rbranch')


class GitReviewUnitTest(testtools.TestCase):
    """Class for misc unit tests."""

    @mock.patch('requests.get', return_value=FakeResponse(404))
    def test_run_http_exc_raise_http_error(self, mock_get):
        url = 'http://gerrit.example.com'
        try:
            cmd.run_http_exc(FakeException, url)
            self.fails('Exception expected')
        except FakeException as err:
            self.assertEqual(cmd.http_code_2_return_code(404), err.code)
            mock_get.assert_called_once_with(url)

    @mock.patch('requests.get', side_effect=Exception())
    def test_run_http_exc_raise_unknown_error(self, mock_get):
        url = 'http://gerrit.example.com'
        try:
            cmd.run_http_exc(FakeException, url)
            self.fails('Exception expected')
        except FakeException as err:
            self.assertEqual(255, err.code)
            mock_get.assert_called_once_with(url)

    @mock.patch('git_review.cmd.run_command_status')
    @mock.patch('requests.get', return_value=FakeResponse(200))
    def test_run_http_exc_without_auth(self, mock_get, mock_run):
        url = 'http://user@gerrit.example.com'

        cmd.run_http_exc(FakeException, url)
        self.assertFalse(mock_run.called)
        mock_get.assert_called_once_with(url)

    @mock.patch('git_review.cmd.run_command_status',
                return_value=(0, FAKE_GIT_CREDENTIAL_FILL))
    @mock.patch('requests.get',
                side_effect=[FakeResponse(401), FakeResponse(200)])
    def test_run_http_exc_with_auth(self, mock_get, mock_run):
        url = 'http://user@gerrit.example.com'

        cmd.run_http_exc(FakeException, url)
        # This gets encoded to utf8 which means the type passed down
        # is bytes.
        mock_run.assert_called_once_with('git', 'credential', 'fill',
                                         stdin='url=%s' % url)
        calls = [mock.call(url), mock.call(url, auth=('user', 'pass'))]
        mock_get.assert_has_calls(calls)

    @mock.patch('git_review.cmd.run_command_status',
                return_value=(0, FAKE_GIT_CREDENTIAL_FILL))
    @mock.patch('requests.get', return_value=FakeResponse(401))
    def test_run_http_exc_with_failing_auth(self, mock_get, mock_run):
        url = 'http://user@gerrit.example.com'

        try:
            cmd.run_http_exc(FakeException, url)
            self.fails('Exception expected')
        except FakeException as err:
            self.assertEqual(cmd.http_code_2_return_code(401), err.code)
        # This gets encoded to utf8 which means the type passed down
        # is bytes.
        mock_run.assert_called_once_with('git', 'credential', 'fill',
                                         stdin='url=%s' % url)
        calls = [mock.call(url), mock.call(url, auth=('user', 'pass'))]
        mock_get.assert_has_calls(calls)

    @mock.patch('git_review.cmd.run_command_status',
                return_value=(1, ''))
    @mock.patch('requests.get', return_value=FakeResponse(401))
    def test_run_http_exc_with_failing_git_creds(self, mock_get, mock_run):
        url = 'http://user@gerrit.example.com'

        try:
            cmd.run_http_exc(FakeException, url)
            self.fails('Exception expected')
        except FakeException as err:
            self.assertEqual(cmd.http_code_2_return_code(401), err.code)
        # This gets encoded to utf8 which means the type passed down
        # is bytes.
        mock_run.assert_called_once_with('git', 'credential', 'fill',
                                         stdin='url=%s' % url)
        mock_get.assert_called_once_with(url)

    @mock.patch('sys.argv', ['argv0', '--track', 'branch'])
    @mock.patch('git_review.cmd.check_remote')
    @mock.patch('git_review.cmd.resolve_tracking')
    def test_command_line_no_track(self, resolve_tracking, check_remote):
        check_remote.side_effect = Exception()
        self.assertRaises(Exception, cmd._main)
        self.assertFalse(resolve_tracking.called)

    @mock.patch('sys.argv', ['argv0', '--track'])
    @mock.patch('git_review.cmd.check_remote')
    @mock.patch('git_review.cmd.resolve_tracking')
    def test_track(self, resolve_tracking, check_remote):
        check_remote.side_effect = Exception()
        self.assertRaises(Exception, cmd._main)
        self.assertTrue(resolve_tracking.called)

    @mock.patch('sys.argv', ['argv0', '-n', '--notify', 'OWNER'])
    @mock.patch('git_review.cmd.check_remote')
    @mock.patch('git_review.cmd.rebase_changes')
    @mock.patch('git_review.cmd.set_hooks_commit_msg')
    @mock.patch('git_review.cmd.assert_one_change')
    @mock.patch('sys.stdout', io.StringIO())
    @mock.patch('sys.exit')
    def test_notify(self, check_remote, rebase, set_hooks, assert_one, exit):
        cmd._main()
        sys.stdout.seek(0)
        stdout = sys.stdout.read()
        self.assertIn("notify=OWNER", stdout)


class DownloadFlagUnitTest(testtools.TestCase):

    def setUp(self):
        super(DownloadFlagUnitTest, self).setUp()
        self.parser = argparse.ArgumentParser()
        self.parser.add_argument(
            '-d',
            action=cmd._DownloadFlag,
            const='download',
            dest='cid',
        )

    def test_store_id(self):
        args = self.parser.parse_args(['-d', '12345'])
        self.assertEqual('12345', args.cid)

    def test_parse_url(self):
        args = self.parser.parse_args(
            ['-d',
             'https://review.opendev.org/12345']
        )
        self.assertEqual('12345', args.cid)

    def test_parse_url_trailing_slash(self):
        args = self.parser.parse_args(
            ['-d',
             'https://review.opendev.org/12345/']
        )
        self.assertEqual('12345', args.cid)

    def test_parse_url_with_update(self):
        args = self.parser.parse_args(
            ['-d',
             'https://review.opendev.org/12345/2']
        )
        self.assertEqual('12345,2', args.cid)

    def test_parse_url_with_hash(self):
        args = self.parser.parse_args(
            ['-d',
             'https://review.opendev.org/#/c/12345']
        )
        self.assertEqual('12345', args.cid)

    def test_parse_url_with_hash_and_update(self):
        args = self.parser.parse_args(
            ['-d',
             'https://review.opendev.org/#/c/12345/1']
        )
        self.assertEqual('12345,1', args.cid)

    def test_parse_polygerrit_url(self):
        args = self.parser.parse_args(
            ['-d',
             'https://review.opendev.org/c/org/project/+/12345']
        )
        self.assertEqual('12345', args.cid)


class RepoUrlParsingTest(testtools.TestCase):
    """Test Git URL parsing"""

    def test_ssh_opendev(self):
        self.assertEqual(
            cmd.parse_gerrit_ssh_params_from_git_url(
                'ssh://someone@review.opendev.org:29418/x/y'),
            ('review.opendev.org', 'someone', '29418',
             'x/y'))

    def test_ssh_no_user_no_port(self):
        self.assertEqual(
            cmd.parse_gerrit_ssh_params_from_git_url(
                'ssh://review.opendev.org/opendev/git-review'),
            ('review.opendev.org', None, None, 'opendev/git-review'))

    def test_ssh_at_sign(self):
        self.assertEqual(
            cmd.parse_gerrit_ssh_params_from_git_url(
                'ssh://someone@example.org@review.opendev.org:29418/x/y'),
            ('review.opendev.org', 'someone@example.org', '29418',
             'x/y'))

    def test_ssh_at_signs_excessive(self):
        self.assertEqual(
            cmd.parse_gerrit_ssh_params_from_git_url(
                'ssh://x@y@example.org@review.opendev.org:29418/x/y'),
            ('review.opendev.org', 'x@y@example.org', '29418',
             'x/y'))

    def test_ssh_at_sign_escaped(self):
        self.assertEqual(
            cmd.parse_gerrit_ssh_params_from_git_url(
                r'ssh://someone%40example.org@review.opendev.org:29418/x/y'),
            ('review.opendev.org', r'someone%40example.org', '29418',
             'x/y'))


@mock.patch('git_review.cmd.run_command')
class GitVersionTest(testtools.TestCase):
    def test_version_too_low(self, mock_run_command):
        # From an old CentOS 7 box
        mock_run_command.return_value = 'git version 1.8.3.1'
        self.assertRaises(SystemExit, cmd.get_git_version)

    def test_version_good(self, mock_run_command):
        mock_run_command.return_value = 'git version 2.43.0'
        self.assertIsNone(cmd.get_git_version())
        self.assertEqual(cmd.LOCAL_GIT_VERSION, (2, 43, 0))



# Copyright (C) 2009 The Android Open Source Project
# Copyright OpenDev Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.  You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Because wheels are unable to embed and install data files, we store
# our vendored copy of the Gerrit commit-msg hook script in a string and
# write it into place at runtime.

# Taken from https://gerrit.googlesource.com/gerrit/+/refs/heads/master
# /resources/com/google/gerrit/server/tools/root/hooks/commit-msg as of
# commit 05d6470 on 2024-02-21.

COMMIT_MSG = r"""#!/bin/sh
#
# Part of Gerrit Code Review (https://www.gerritcodereview.com/)
#
# Copyright (C) 2009 The Android Open Source Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

set -u

# avoid [[ which is not POSIX sh.
if test "$#" != 1 ; then
  echo "$0 requires an argument."
  exit 1
fi

if test ! -f "$1" ; then
  echo "file does not exist: $1"
  exit 1
fi

# Do not create a change id if requested
case "$(git config --get gerrit.createChangeId)" in
  false)
    exit 0
    ;;
  always)
    ;;
  *)
    # Do not create a change id for squash/fixup commits.
    if head -n1 "$1" | LC_ALL=C grep -q '^[a-z][a-z]*! '; then
      exit 0
    fi
    ;;
esac


if git rev-parse --verify HEAD >/dev/null 2>&1; then
  refhash="$(git rev-parse HEAD)"
else
  refhash="$(git hash-object -t tree /dev/null)"
fi

random=$({ git var GIT_COMMITTER_IDENT ; echo "$refhash" ; cat "$1"; } | git hash-object --stdin)
dest="$1.tmp.${random}"

trap 'rm -f "$dest" "$dest-2"' EXIT

if ! cat "$1" | sed -e '/>8/q' | git stripspace --strip-comments > "${dest}" ; then
   echo "cannot strip comments from $1"
   exit 1
fi

if test ! -s "${dest}" ; then
  echo "file is empty: $1"
  exit 1
fi

reviewurl="$(git config --get gerrit.reviewUrl)"
if test -n "${reviewurl}" ; then
  token="Link"
  value="${reviewurl%/}/id/I$random"
  pattern=".*/id/I[0-9a-f]\{40\}"
else
  token="Change-Id"
  value="I$random"
  pattern=".*"
fi

if git interpret-trailers --parse < "$1" | grep -q "^$token: $pattern$" ; then
  exit 0
fi

# There must be a Signed-off-by trailer for the code below to work. Insert a
# sentinel at the end to make sure there is one.
# Avoid the --in-place option which only appeared in Git 2.8
if ! git interpret-trailers \
         --trailer "Signed-off-by: SENTINEL" < "$1" > "$dest-2" ; then
  echo "cannot insert Signed-off-by sentinel line in $1"
  exit 1
fi

# Make sure the trailer appears before any Signed-off-by trailers by inserting
# it as if it was a Signed-off-by trailer and then use sed to remove the
# Signed-off-by prefix and the Signed-off-by sentinel line.
# Avoid the --in-place option which only appeared in Git 2.8
# Avoid the --where option which only appeared in Git 2.15
if ! git -c trailer.where=before interpret-trailers \
         --trailer "Signed-off-by: $token: $value" < "$dest-2" |
     sed -e "s/^Signed-off-by: \($token: \)/\1/" \
         -e "/^Signed-off-by: SENTINEL/d" > "$dest" ; then
  echo "cannot insert $token line in $1"
  exit 1
fi

if ! mv "${dest}" "$1" ; then
  echo "cannot mv ${dest} to $1"
  exit 1
fi
"""  # noqa



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import hashlib
import os
import shutil
import stat
import struct
import sys

import fixtures
import requests
import testtools
from testtools import content

from git_review.hooks import COMMIT_MSG
from git_review.tests import utils

if sys.version < '3':
    import urllib
    import urlparse
    urlparse = urlparse.urlparse
else:
    import urllib.parse
    import urllib.request
    urlparse = urllib.parse.urlparse


DEFAULT_WAR_URL = 'https://gerrit-releases.storage.googleapis.com/' \
                  'gerrit-3.4.4.war'
# Update GOLDEN_SITE_VER for every change altering golden site, including
# DEFAULT_WAR_URL changes. Set new value to something unique (just +1 it for
# example).
GOLDEN_SITE_VER = '7'


# NOTE(yorik-sar): This function needs to be a perfect hash function for
# existing test IDs. This is verified by running check_test_id_hashes script
# prior to running tests. Note that this doesn't imply any cryptographic
# requirements for underlying algorithm, so we can use weak hash here.
# Range of results for this function is limited by port numbers selection
# in _pick_gerrit_port_and_dir method (it can't be greater than 10000 now).
def _hash_test_id(test_id):
    if not isinstance(test_id, bytes):
        test_id = test_id.encode('utf-8')
    hash_ = hashlib.md5(test_id).digest()
    num = struct.unpack("=I", hash_[:4])[0]
    return num % 10000


class DirHelpers(object):
    def _dir(self, base, *args):
        """Creates directory name from base name and other parameters."""
        return os.path.join(getattr(self, base + '_dir'), *args)


class IsoEnvDir(DirHelpers, fixtures.Fixture):
    """Created isolated env and associated directories"""

    def __init__(self, base_dir=None):
        super(IsoEnvDir, self).setUp()

        # set up directories for isolation
        if base_dir:
            self.base_dir = base_dir
            self.temp_dir = self._dir('base', 'tmp')
            if not os.path.exists(self.temp_dir):
                os.mkdir(self.temp_dir)
            self.addCleanup(shutil.rmtree, self.temp_dir)
        else:
            self.temp_dir = self.useFixture(fixtures.TempDir()).path

        self.work_dir = self._dir('temp', 'work')
        self.home_dir = self._dir('temp', 'home')
        self.xdg_config_dir = self._dir('home', '.xdgconfig')
        for path in [self.home_dir, self.xdg_config_dir, self.work_dir]:
            if not os.path.exists(path):
                os.mkdir(path)

        # ensure user proxy conf doesn't interfere with tests
        self.useFixture(fixtures.EnvironmentVariable('no_proxy', '*'))
        self.useFixture(fixtures.EnvironmentVariable('NO_PROXY', '*'))

        # isolate tests from user and system git configuration
        self.useFixture(fixtures.EnvironmentVariable('HOME', self.home_dir))
        self.useFixture(
            fixtures.EnvironmentVariable('XDG_CONFIG_HOME',
                                         self.xdg_config_dir))
        self.useFixture(
            fixtures.EnvironmentVariable('GIT_CONFIG_NOSYSTEM', "1"))
        self.useFixture(
            fixtures.EnvironmentVariable('EMAIL', "you@example.com"))
        self.useFixture(
            fixtures.EnvironmentVariable('GIT_AUTHOR_NAME',
                                         "gitreview tester"))
        self.useFixture(
            fixtures.EnvironmentVariable('GIT_COMMITTER_NAME',
                                         "gitreview tester"))


class GerritHelpers(DirHelpers):
    def init_dirs(self):
        self.primary_dir = os.path.abspath(os.path.curdir)
        self.gerrit_dir = self._dir('primary', '.gerrit')
        self.gsite_dir = self._dir('gerrit', 'golden_site')
        self.war_url = os.environ.get('WAR_URL', DEFAULT_WAR_URL)
        self.gerrit_war = self._dir('gerrit', self.war_url.split('/')[-1])

    def ensure_gerrit_war(self):
        # check if gerrit.war file exists in .gerrit directory
        if not os.path.exists(self.gerrit_dir):
            os.mkdir(self.gerrit_dir)

        if not os.path.exists(self.gerrit_war):
            print("Downloading Gerrit binary from %s..." % self.war_url)
            resp = requests.get(self.war_url)
            if resp.status_code != 200:
                raise RuntimeError("Problem requesting Gerrit war")
            utils.write_to_file(self.gerrit_war, resp.content)
            print("Saved to %s" % self.gerrit_war)

    def init_gerrit(self):
        """Run Gerrit from the war file and configure it."""
        golden_ver_file = self._dir('gsite', 'golden_ver')
        if os.path.exists(self.gsite_dir):
            if not os.path.exists(golden_ver_file):
                golden_ver = '0'
            else:
                with open(golden_ver_file) as f:
                    golden_ver = f.read().strip()
            if GOLDEN_SITE_VER != golden_ver:
                print("Existing golden site has version %s, removing..." %
                      golden_ver)
                shutil.rmtree(self.gsite_dir)
            else:
                print("Golden site of version %s already exists" %
                      GOLDEN_SITE_VER)
                return

        print("Creating a new golden site of version " + GOLDEN_SITE_VER)

        # initialize Gerrit in developer mode, and include the
        # download-commands plugin (needed by some tests which try to retrieve
        # by Change-Id string)
        utils.run_cmd('java', '-jar', self.gerrit_war,
                      'init', '-d', self.gsite_dir, '--dev',
                      '--batch', '--no-auto-start', '--install-plugin',
                      'download-commands')

        # pre-index so that it won't be triggered on start for each copy
        utils.run_cmd('java', '-jar', self.gerrit_war, 'reindex',
                      '-d', self.gsite_dir)

        # record our site version
        with open(golden_ver_file, 'w') as f:
            f.write(GOLDEN_SITE_VER)

        # Create SSH public key. This is not RSA because Jammy wants to
        # do rsa + sha2 by default and Gerrit 3.4.4 does not understand this.
        utils.run_cmd('ssh-keygen', '-t', 'ed25519',
                      '-f', self._dir('gsite', 'test_ssh_key'), '-N', '')

    def _run_gerrit_cli(self, command, *args):
        """SSH to gerrit Gerrit server and run command there."""
        return utils.run_cmd('ssh', '-p', str(self.gerrit_port),
                             'admin@' + self.gerrit_host, 'gerrit',
                             command, *args)

    def _run_git_review(self, *args, **kwargs):
        """Run git-review utility from source."""
        git_review = utils.run_cmd('which', 'git-review')
        kwargs.setdefault('chdir', self.test_dir)
        return utils.run_cmd(git_review, *args, **kwargs)


class BaseGitReviewTestCase(testtools.TestCase, GerritHelpers):
    """Base class for the git-review tests."""

    _remote = 'gerrit'

    @property
    def project_uri(self):
        return self.project_ssh_uri

    def setUp(self):
        """Configure testing environment.

        Prepare directory for the testing and clone test Git repository.
        Require Gerrit war file in the .gerrit directory to run Gerrit local.
        """
        super(BaseGitReviewTestCase, self).setUp()
        self.useFixture(fixtures.Timeout(5 * 60, True))

        # ensures git-review command runs in local mode (for functional tests)
        self.useFixture(
            fixtures.EnvironmentVariable('GITREVIEW_LOCAL_MODE', ''))

        self.init_dirs()
        ssh_addr, ssh_port, http_addr, http_port, self.site_dir = \
            self._pick_gerrit_port_and_dir()
        self.gerrit_host, self.gerrit_port = ssh_addr, ssh_port

        self.test_dir = self._dir('site', 'tmp', 'test_project')
        self.ssh_dir = self._dir('site', 'tmp', 'ssh')
        self.project_ssh_uri = (
            'ssh://admin@%s:%s/test/test_project.git' % (
                ssh_addr, ssh_port))
        self.project_http_uri = (
            'http://admin:secret@%s:%s/test/test_project.git' % (
                http_addr, http_port))

        self._run_gerrit(ssh_addr, ssh_port, http_addr, http_port)
        self._configure_ssh(ssh_addr, ssh_port)

        # Upload SSH key for test admin user
        with open(self._dir('gsite', 'test_ssh_key.pub'), 'rb') as pub_key_fd:
            pub_key = pub_key_fd.read().decode().strip()
        resp = requests.post(
            'http://%s:%s/a/accounts/self/sshkeys' % (http_addr, http_port),
            auth=requests.auth.HTTPBasicAuth('admin', 'secret'),
            headers={'Content-Type': 'text/plain'},
            data=pub_key)
        if resp.status_code != 201:
            raise RuntimeError(
                'SSH key upload failed: %s "%s"' % (resp, resp.text))

        # create Gerrit empty project
        self._run_gerrit_cli('create-project', 'test/test_project',
                             '--empty-commit')

        # setup isolated area to work under
        self.useFixture(IsoEnvDir(self.site_dir))

        # prepare repository for the testing
        self._run_git('clone', self.project_uri)
        utils.write_to_file(self._dir('test', 'test_file.txt'),
                            'test file created'.encode())
        self._create_gitreview_file()
        hook_file = self._dir('test', '.git/hooks/commit-msg')
        utils.write_to_file(hook_file, COMMIT_MSG.encode())
        os.chmod(hook_file, os.stat(hook_file).st_mode | stat.S_IEXEC)

        # push changes to the Gerrit
        self._run_git('add', '--all')
        self._run_git('commit', '-m', 'Test file and .gitreview added.')
        self._run_git('push', 'origin', 'HEAD:refs/for/master')
        head = self._run_git('rev-parse', 'HEAD')
        self._run_gerrit_cli('review', head, '--code-review=+2', '--submit')
        # push a branch to gerrit
        self._run_gerrit_cli('create-branch', 'test/test_project',
                             'testbranch', head)
        self._run_git('checkout', '-b', 'testbranch')
        utils.write_to_file(self._dir('test', 'test_file.txt'),
                            'test file branched'.encode())
        self._create_gitreview_file(defaultbranch='testbranch')
        self._run_git('add', '--all')
        self._run_git('commit', '-m', 'Branched.')
        self._run_git('push', 'origin', 'HEAD:refs/for/testbranch')
        head = self._run_git('rev-parse', 'HEAD')
        self._run_gerrit_cli('review', head, '--code-review=+2', '--submit')
        # cleanup. This is important so our work above doesn't pollute the
        # test cases. We need git review to set stuff up like commit message
        # hooks.
        shutil.rmtree(self.test_dir)

        # go to the just cloned test Git repository
        self._run_git('clone', self.project_uri)
        self.configure_gerrit_remote()
        self.addCleanup(shutil.rmtree, self.test_dir)

        # ensure user is configured for all tests
        self._configure_gitreview_username()

    def set_remote(self, uri):
        self._run_git('remote', 'set-url', self._remote, uri)

    def reset_remote(self):
        self._run_git('remote', 'rm', self._remote)

    def attach_on_exception(self, filename):
        @self.addOnException
        def attach_file(exc_info):
            if os.path.exists(filename):
                content.attach_file(self, filename)
            else:
                self.addDetail(os.path.basename(filename),
                               content.text_content('Not found'))

    def _run_git(self, command, *args):
        """Run git command using test git directory."""
        if command == 'clone':
            return utils.run_git(command, args[0], self._dir('test'))
        return utils.run_git('--git-dir=' + self._dir('test', '.git'),
                             '--work-tree=' + self._dir('test'),
                             command, *args)

    def _run_git_sub(self, command, *args):
        """Run git command using submodule of test git directory."""
        if command == 'init':
            utils.run_git('init', self._dir('test', 'sub'))
            self._simple_change_sub('submodule content', 'initial commit')
            utils.run_git('submodule', 'add', os.path.join('.', 'sub'),
                          chdir=self._dir('test'))
            return self._run_git('commit', '-m', 'add submodule')
        return utils.run_git('--git-dir=' + self._dir('test', 'sub', '.git'),
                             '--work-tree=' + self._dir('test', 'sub'),
                             command, *args)

    def _run_gerrit(self, ssh_addr, ssh_port, http_addr, http_port):
        # create a copy of site dir
        if os.path.exists(self.site_dir):
            shutil.rmtree(self.site_dir)
        shutil.copytree(self.gsite_dir, self.site_dir)
        self.addCleanup(shutil.rmtree, self.site_dir)
        # write config
        old_conf = configparser.ConfigParser(strict=False)
        old_conf.read(self._dir('site', 'etc', 'gerrit.config'))
        with open(self._dir('site', 'etc', 'gerrit.config'), 'w') as _conf:
            new_conf = utils.get_gerrit_conf(
                ssh_addr, ssh_port, http_addr, http_port)
            _conf.write(new_conf)
        # Preserve the preexisting javaHome configuration
        new_conf = configparser.ConfigParser(strict=False)
        new_conf.read(self._dir('site', 'etc', 'gerrit.config'))
        new_conf['container'] = {}
        new_conf['container']['javaHome'] = old_conf['container']['javaHome']
        with open(self._dir('site', 'etc', 'gerrit.config'), 'w') as _conf:
            new_conf.write(_conf)

        # If test fails, attach Gerrit config and logs to the result
        self.attach_on_exception(self._dir('site', 'etc', 'gerrit.config'))
        for name in ['error_log', 'sshd_log', 'httpd_log']:
            self.attach_on_exception(self._dir('site', 'logs', name))

        # start Gerrit
        gerrit_sh = self._dir('site', 'bin', 'gerrit.sh')
        out = utils.run_cmd(gerrit_sh, 'start')
        self.addDetail("gerrit-start-output",
                       content.text_content(out))
        self.addCleanup(utils.run_cmd, gerrit_sh, 'stop')

    def _unstaged_change(self, change_text, file_=None):
        """Helper method to create small changes and not stage them."""
        if file_ is None:
            file_ = self._dir('test', 'test_file.txt')
        utils.write_to_file(file_, ''.encode())
        self._run_git('add', file_)
        utils.write_to_file(file_, change_text.encode())

    def _uncommitted_change(self, change_text, file_=None):
        """Helper method to create small changes and not commit them."""
        if file_ is None:
            file_ = self._dir('test', 'test_file.txt')
        self._unstaged_change(change_text, file_)
        self._run_git('add', file_)

    def _simple_change(self, change_text, commit_message,
                       file_=None):
        """Helper method to create small changes and commit them."""
        self._uncommitted_change(change_text, file_)
        self._run_git('commit', '-m', commit_message)

    def _simple_amend(self, change_text, file_=None):
        """Helper method to amend existing commit with change."""
        if file_ is None:
            file_ = self._dir('test', 'test_file_new.txt')
        utils.write_to_file(file_, change_text.encode())
        self._run_git('add', file_)
        # cannot use --no-edit because it does not exist in older git
        message = self._run_git('log', '-1', '--format=%s\n\n%b')
        self._run_git('commit', '--amend', '-m', message)

    def _unstaged_change_sub(self, change_text, file_=None):
        """Helper method to create small submodule changes and not stage."""
        if file_ is None:
            file_ = self._dir('test', 'sub', 'test_file.txt')
        utils.write_to_file(file_, ''.encode())
        self._run_git_sub('add', file_)
        utils.write_to_file(file_, change_text.encode())

    def _uncommitted_change_sub(self, change_text, file_=None):
        """Helper method to create small submodule changes and not commit."""
        if file_ is None:
            file_ = self._dir('test', 'sub', 'test_file.txt')
        self._unstaged_change_sub(change_text, file_)
        self._run_git_sub('add', file_)

    def _simple_change_sub(self, change_text, commit_message, file_=None):
        """Helper method to create small submodule changes and commit them."""
        self._uncommitted_change_sub(change_text, file_)
        self._run_git_sub('commit', '-m', commit_message)

    def _configure_ssh(self, ssh_addr, ssh_port):
        """Setup ssh and scp to run with special options."""

        os.mkdir(self.ssh_dir)

        ssh_key = utils.run_cmd('ssh-keyscan', '-p', str(ssh_port), ssh_addr)
        utils.write_to_file(self._dir('ssh', 'known_hosts'), ssh_key.encode())
        self.addCleanup(os.remove, self._dir('ssh', 'known_hosts'))

        # Attach known_hosts to test results if anything fails
        self.attach_on_exception(self._dir('ssh', 'known_hosts'))

        for cmd in ('ssh', 'scp'):
            cmd_file = self._dir('ssh', cmd)
            s = '#!/bin/sh\n' \
                '/usr/bin/%s -i %s -o UserKnownHostsFile=%s ' \
                '-o IdentitiesOnly=yes ' \
                '-o PasswordAuthentication=no $@' % \
                (cmd,
                 self._dir('gsite', 'test_ssh_key'),
                 self._dir('ssh', 'known_hosts'))
            utils.write_to_file(cmd_file, s.encode())
            os.chmod(cmd_file, os.stat(cmd_file).st_mode | stat.S_IEXEC)

        os.environ['PATH'] = self.ssh_dir + os.pathsep + os.environ['PATH']
        os.environ['GIT_SSH'] = self._dir('ssh', 'ssh')

    def configure_gerrit_remote(self):
        self._run_git('remote', 'add', self._remote, self.project_uri)

    def _configure_gitreview_username(self):
        self._run_git('config', 'gitreview.username', 'admin')

    def _pick_gerrit_port_and_dir(self):
        hash_ = _hash_test_id(self.id())
        host = '127.0.0.1'
        return (
            host, 12000 + hash_,  # avoid 11211 that is memcached port on CI
            host, 22000 + hash_,  # avoid ephemeral ports at 32678+
            self._dir('gerrit', 'site-' + str(hash_)),
        )

    def _create_gitreview_file(self, **kwargs):
        cfg = ('[gerrit]\n'
               'scheme=%s\n'
               'host=%s\n'
               'port=%s\n'
               'project=test/test_project.git\n'
               '%s')
        parsed = urlparse(self.project_uri)
        host_port = parsed.netloc.rpartition('@')[-1]
        host, __, port = host_port.partition(':')
        extra = '\n'.join('%s=%s' % kv for kv in kwargs.items())
        cfg %= parsed.scheme, host, port, extra
        utils.write_to_file(self._dir('test', '.gitreview'), cfg.encode())


class HttpMixin(object):
    """HTTP remote_url mixin."""

    @property
    def project_uri(self):
        return self.project_http_uri

    def _configure_gitreview_username(self):
        # trick to set http password
        self._run_git('config', 'gitreview.username', 'admin:secret')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.  You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import nox


nox.options.error_on_external_run = True
nox.options.reuse_existing_virtualenvs = True
nox.options.sessions = ["tests-3", "docs", "linters"]


@nox.session(python="3")
def linters(session):
    session.install("-r", "requirements.txt")
    session.install("-r", "test-requirements.txt")
    session.install("-e", ".")
    session.run("flake8")


@nox.session(python="3")
def docs(session):
    session.install("-r", "requirements.txt")
    session.install("-r", "doc/requirements.txt")
    session.install("-e", ".")
    session.run(
        "sphinx-build", "-W",
        "-d", "doc/build/doctrees",
        "-b", "html",
        "doc/source/", "doc/build/html"
    )


@nox.session(python="3")
def venv(session):
    session.install("-r", "requirements.txt")
    session.install("-r", "test-requirements.txt")
    session.install("-e", ".")
    session.run(*session.posargs)


@nox.session(python="3")
def tests(session):
    session.install("-r", "requirements.txt")
    session.install("-r", "test-requirements.txt")
    session.install("-e", ".")
    session.run("python", "-m", "git_review.tests.check_test_id_hashes",
                "discover", "--list")
    session.run("python", "-m", "git_review.tests.prepare")
    session.run("stestr", "run", "--color", *session.posargs)
